// E.I.S. / ESA modlib for COMRADE and HardScale
  //
  //// Description:
  // ----------------
  // Universal bitselect operation
  // with output queue and shift register mechanism
  //
  // Version History:
  // ----------------
  // 100826: BT: intial release

`timescale 1ns / 1ns

module bitsel
         #(parameter
                      WA                   = 32,
                      WR                   = 32,
                      DEPTH                = 0,
                      QDEPTH               = 16,
                      TQDEPTH              = 16,
                      STATIC_CT            = 0,
                      START_CTRL_IN        = 0,
                      DATA_SPECULATION     = 0,
                      SPECULATION_BOUNDARY = 0,
                      PSEUDO_AT            = 0,
                      SIGN                 = 0
          )(
           input  wire    [(WA-1):0] A,
           output wire    [(WR-1):0] R,
           input  wire               RESET,
           input  wire               CLK,
           input  wire               CE,
           input  wire               CONTROLLER_CANCEL,
           output wire               TOKENS_EMPTY,
           output wire               CURRENT_TOKEN,
           output wire               CURRENT_TOKEN_PSEUDO,
           input  wire               START,
           output wire               START_ACK,
           input  wire               START_CTRL,
           output wire               START_CTRL_ACK,
           input  wire               START_CTRL_PSEUDO,
           input  wire               CANCEL,
           output reg                CANCEL_ACK,
           output wire               CANCEL_STATE,
           input  wire               CANCEL_STATE_ACK,
           output wire               CANCEL_STATE_CTRL,
           input  wire               CANCEL_STATE_CTRL_ACK,
           output wire               RESULT_READY,
           input  wire               RESULT_READY_ACK);

  // define LATENCY of this module
  localparam LATENCY          = 1;
  // CANCEL BITS depending on LATENCY and shift register DEPTH
  localparam CANCEL_BITS      = min_1(ceil_log2(LATENCY+DEPTH-1));
  // QDEPTH_INT has to be at least 16
  localparam QDEPTH_INT       = min_16(QDEPTH);
  // set QBITS accordingly
  localparam QBITS            = (QDEPTH == 0) ? 0 : ceil_log2(QDEPTH_INT-1);
  // TQDEPTH_INT has to be at least 16
  localparam TQDEPTH_INT      = min_16(TQDEPTH);
  // set TQBITS accordingly
  localparam TQBITS           = (TQDEPTH == 0) ? 0 : ceil_log2(TQDEPTH_INT-1);
  // COMPUTATION_BITS depending  on LATENCY shift register DEPTH and queue QDEPTH
  localparam COMPUTATION_BITS = min_1(ceil_log2(LATENCY+DEPTH+QDEPTH_INT-1));


  wire [(WR-1)              :0]    r_int;

  `include "handshaking_assigns_verilog_only.v"

  generate
  if (WA < WR) begin
    assign r_int = ((SIGN==1) ? {{(WR-WA){A[WA-1]}}, A[(WA-1):0]} : {{(WR-WA){1'b0}}, A[(WA-1):0]});
  end
  else if (WA == WR) begin
    assign r_int = ((SIGN==1) ? {A[(WA-1):0]} : {A[(WA-1):0]});
  end
  else begin
    wire [(WR-1):0] neg_signed;
    wire [(WR-1):0] pos_signed;
    wire [(WR-1):0] abs_unsigned;
    // the operation
    assign neg_signed   = &A[(WA-1):(WR-1)] ? A[(WR-1):0] : {1'b1, {(WR-1){1'b0}}};   // signed negative out of range?
    assign pos_signed   = |A[(WA-1):(WR-1)] ? {1'b0, {(WR-1){1'b1}}} : A[(WR-1):0];   // signed positive out of range?
    assign abs_unsigned = |A[(WA-1):(WR)]   ? {(WR){1'b1}} : A[(WR-1):0];             // unsigned out of range?
    assign r_int = ((SIGN==1) ? A[(WA-1)] ? neg_signed : pos_signed : abs_unsigned);
  end
  endgenerate

  // shift register instantiation
  shiftreg #(.WA(WR), .WR(WR), .DEPTH(DEPTH)) sr (
    .A(r_int),
    .R(DATA_SHIFTREG_TO_QUEUE),
    .RESET(RESET),
    .CLK(CLK),
    .CE(CE_SHIFTREG),
    .START(START_SHIFTREG),
    .RESULT_READY(RESULT_READY_SHIFTREG));


  `include "functions.v"

endmodule
