// E.I.S. / ESA modlib for COMRADE and HardScale
  //
  //// Description:
  // ----------------
  // Const module - emitting always the same value, which is configured via
  // a parameter. Cancel tokens just vanish in this node, because it's a
  // start of a data path. For making the node capable of outputting tokens
  // each cycle, it an AT-Counter which size is configurable with parameter QDEPTH.
  // I takes 1 cycle to activate the constant (i.e. ATs do not move through transparently.)
  //
  // Version History:
  // ----------------
  // 100826: BT: intial release

`timescale 1ns / 1ns

module const_op
         #(parameter
                      WR                   = 32,           // 1 for logic operations
                      SIGN                 = 0,
                      VALUE                = 32'hDEADBEEF, // the constant output data value
                      DEPTH                = 0,            // dummy
                      STATIC_CT            = 0,
                      START_CTRL_IN        = 0,
                      DATA_SPECULATION     = 0,
                      SPECULATION_BOUNDARY = 0,
                      PSEUDO_AT            = 0,
                      TQDEPTH              = 0,
                      QDEPTH               = 1
          )(
           output wire    [(WR-1):0] R,
           input  wire               RESET,
           input  wire               CLK,
           input  wire               CE,
           input  wire               CONTROLLER_CANCEL,
           output wire               TOKENS_EMPTY,
           output wire               CURRENT_TOKEN,
           output wire               CURRENT_TOKEN_PSEUDO,
           input  wire               START,
           output wire               START_ACK,
           input  wire               START_CTRL,
           output wire               START_CTRL_ACK,
           input  wire               START_CTRL_PSEUDO,
           input  wire               CANCEL,
           output wire               CANCEL_ACK,
           output wire               CANCEL_STATE,
           input  wire               CANCEL_STATE_ACK,
           output wire               CANCEL_STATE_CTRL,
           input  wire               CANCEL_STATE_CTRL_ACK,
           output wire               RESULT_READY,
           input  wire               RESULT_READY_ACK);

  // configure size of AT-counter - must always be >= 1
  localparam QDEPTH_NZ = min_1(QDEPTH);

  wire  increase_at_counter;
  
  wire  computation_counter = 1'b0;
  wire  cancel_counter = 1'b0;
 
  reg [QDEPTH_NZ-1:0] AT_COUNTER;
  
  generate
  if (DATA_SPECULATION == 0) begin
    assign  increase_at_counter = START & START_ACK & START_CTRL & START_CTRL_ACK;
  
  end
  endgenerate
  
  wire decrease_at_counter = RESULT_READY & RESULT_READY_ACK; // RESULT_READY needed to not decrement below 0  
    
  assign TOKENS_EMPTY = 1'b0;
  assign CURRENT_TOKEN = 1'b0;
  assign CURRENT_TOKEN_PSEUDO = 1'b0;// value cannot be less than 1  

  // always output constant value
  assign R = VALUE;

  // ATs are accepted as long as there is space in the counter
  generate
  if (DATA_SPECULATION == 0) begin
  assign START_ACK = START & ~(&(AT_COUNTER));
  end
  endgenerate 
  
  assign START_CTRL_ACK = START_CTRL & ~(&(AT_COUNTER));

  // throw away CTs
  assign CANCEL_ACK = 1'b1;
  assign CANCEL_STATE = 1'b0;
  assign CANCEL_STATE_CTRL = 1'b0;

  
  // AT in counter?
  assign RESULT_READY = |(AT_COUNTER);
  
  // AT_COUNTER
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      AT_COUNTER <= {(QDEPTH_NZ){1'b0}};
    end else begin
      if (increase_at_counter && ~decrease_at_counter) begin
        AT_COUNTER <= AT_COUNTER + 1'b1;
      end else if (~increase_at_counter && decrease_at_counter) begin
        AT_COUNTER <= AT_COUNTER - 1'b1;
      end else begin
        AT_COUNTER <= AT_COUNTER;
      end
    end
  end


  `include "functions.v"

endmodule
