// E.I.S. / ESA modlib for COMRADE and HardScale
  //
  // Description:
  // ----------------
  // Output queue
  // based on shift registers
  //
  // Version History:
  // ----------------
  // 100826: BT: intial release


module disram_queue_non_spec
         #(parameter
                      WR                  = 32,
                      QDEPTH              = 16,
                      QBITS               =  4 // 4 bit -> 16 words in the QUEUE.
          )(
           input                     CLK,
           input                     RESET,
           input  [(WR-1):0]         DIN,
           input                     READ,
           input                     WRITE,
           output wire [(WR-1):0]    DOUT,
           output reg                EMPTY,
           output reg                FULL);


 // QBITS==0 can only result from QDEPTH=0; in this case, ADDRESS_SRL
 // is not generated anyway, but XST does not recognize this, so
 // we have to guarantee that in [x:y] for ADDRESS_SRL, x is
 // always >= 0. => Just insert "+(QBITS==0)".
 localparam QBITS_NZ = QBITS + (QBITS==0);

 wire [(WR-1)   :0]            DOUT_SDP;

 reg [(QBITS_NZ-1):0] ADDRESS_SDP_RD;
 reg [(QBITS_NZ-1):0] ADDRESS_SDP_WR;



 assign DOUT                = (QDEPTH == 0) ? DIN : DOUT_SDP;
 
 // generate the required FIFO
  genvar i;

  generate
  if (QDEPTH != 0) begin

  // shift register instantiation
     distr_sdpram #(.DATA_WIDTH(WR), .ADDR_BITS(QBITS_NZ)) DISTSDP_i (
      .CLK(CLK),
      .DIN(DIN),
      .DOUT(DOUT_SDP),
      .ADDR_RD(ADDRESS_SDP_RD[(QBITS_NZ-1):0]),
      .ADDR_WR(ADDRESS_SDP_WR[(QBITS_NZ-1):0]),
      .WE(WRITE));
  end
  endgenerate


generate
if (QDEPTH != 0) begin// address pointer logic
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      ADDRESS_SDP_RD <= {(QBITS_NZ){1'b0}};
      ADDRESS_SDP_WR <= {(QBITS_NZ){1'b0}};
    end else begin
      if (READ == 1'b1 & (EMPTY == 1'b0 | WRITE == 1'b1)) begin
           ADDRESS_SDP_RD <= ADDRESS_SDP_RD + 1'b1;
      end
      else begin
          ADDRESS_SDP_RD <= ADDRESS_SDP_RD;
      end

      // write into FIFO
      if (WRITE == 1'b1 & FULL == 1'b0) begin
           // increase pointer
           ADDRESS_SDP_WR <= ADDRESS_SDP_WR + 1'b1;
      end
      else begin
          ADDRESS_SDP_WR <= ADDRESS_SDP_WR;
      end
    end
  end


  // Comparator Block
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      FULL <= 1'b0;
    end else begin
      if ((ADDRESS_SDP_WR[(QBITS_NZ-1):0] + 1'b1 == ADDRESS_SDP_RD[(QBITS_NZ-1):0]) & (WRITE & ~READ)) begin
        FULL <= 1'b1;
      end
      else if ((ADDRESS_SDP_WR[(QBITS_NZ-1):0] == ADDRESS_SDP_RD[(QBITS_NZ-1):0]) & (~READ | WRITE)) begin
        FULL <= FULL;
      end
      else begin
        FULL <= 1'b0;
      end
    end
  end


  // Empty flag is generated when reading from the last location
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      EMPTY <= 1'b1;
    end else begin
      if ((ADDRESS_SDP_RD[(QBITS_NZ-1):0] + 1'b1 == ADDRESS_SDP_WR[(QBITS_NZ-1):0]) & (READ & ~WRITE)) begin
        EMPTY <= 1'b1;
      end
      else if ((ADDRESS_SDP_RD[(QBITS_NZ-1):0] == ADDRESS_SDP_WR[(QBITS_NZ-1):0]) & (~WRITE | READ)) begin
        EMPTY <= EMPTY;
      end
      else begin
        EMPTY <= 1'b0;
      end
    end
  end

end


else begin
  // Comparator Block
  always @(posedge CLK) begin
    FULL  <= 1'b1;
    EMPTY <= 1'b1;
  end
end
endgenerate

endmodule
