// E.I.S. / ESA modlib for COMRADE and HardScale
  //
  //// Description:
  // ----------------
  // Universal integer division operation
  // with output queue and shift register mechanism
  //
  // Version History:
  // ----------------
  // 100826: BT: intial release

`timescale 1ns / 1ns

`ifdef QUOTIENT
  `undef QUOTIENT
`endif
`ifdef REMAINDER
  `undef REMAINDER
`endif

/*
`ifdef TARGET_ML507
  `define QUOTIENT quotient
  `define REMAINDER fractional
`else*/
  `define QUOTIENT quot
  `define REMAINDER remd
  //`endif

module divint
         #(parameter
                      WA                   = 32,
                      WB                   = 32,
                      WR                   = 32,
                      DEPTH                = 0,
                      QDEPTH               = 16,
                      TQDEPTH              = 16,
                      STATIC_CT            = 0,
                      START_CTRL_IN        = 0,
                      DATA_SPECULATION     = 0,
                      SPECULATION_BOUNDARY = 0,
                      PSEUDO_AT            = 0,
                      AREA_N_SPEED         = 0,
                      SIGN                 = 0
          )(
           input  wire    [(WA-1):0] A,
           input  wire    [(WB-1):0] B,
           output wire    [(WR-1):0] R,
           input  wire               RESET,
           input  wire               CLK,
           input  wire               CE,
           input  wire               CONTROLLER_CANCEL,
           output wire               TOKENS_EMPTY,
           output wire               CURRENT_TOKEN,
           output wire               CURRENT_TOKEN_PSEUDO,
           input  wire               START,
           output wire               START_ACK,
           input  wire               START_CTRL,
           output wire               START_CTRL_ACK,
           input  wire               START_CTRL_PSEUDO,
           input  wire               CANCEL,
           output reg                CANCEL_ACK,
           output wire               CANCEL_STATE,
           input  wire               CANCEL_STATE_ACK,
           output wire               CANCEL_STATE_CTRL,
           input  wire               CANCEL_STATE_CTRL_ACK,
           output wire               RESULT_READY,
           input  wire               RESULT_READY_ACK);

  // define LATENCY of this module (operator only)
  localparam LATENCY          = WA+2+SIGN*2+AREA_N_SPEED; // 32b mod: 34 clocks + 2 clocks if signed + 1 if II>1
  // CANCEL BITS depending on LATENCY and shift register DEPTH
  localparam CANCEL_BITS      = min_1(ceil_log2(LATENCY+DEPTH-1));
  // QDEPTH_INT has to be at least 16
  localparam QDEPTH_INT       = min_16(QDEPTH);
  // set QBITS accordingly
  localparam QBITS            = (QDEPTH == 0) ? 0 : ceil_log2(QDEPTH_INT-1);
  // TQDEPTH_INT has to be at least 16
  localparam TQDEPTH_INT      = min_16(TQDEPTH);
  // set TQBITS accordingly
  localparam TQBITS           = (TQDEPTH == 0) ? 0 : ceil_log2(TQDEPTH_INT-1);
  // COMPUTATION_BITS depending  on LATENCY shift register DEPTH and queue QDEPTH
  localparam COMPUTATION_BITS = min_1(ceil_log2(LATENCY+DEPTH+QDEPTH_INT-1));


  wire [(WR-1):0]       quot;
  reg  [(LATENCY-1):0]  RESULT_READY_int;

  `include "handshaking_assigns_blackbox_include.v"

  // synthesis translate_off
  initial begin
    if (((WA != 32) && (WA != 16) && (WA != 8)) || (WB != WA) || (WR != WA)) begin
      $write("Module DIVINT: Currently supports only a width of 32b/16b/8b for A, B, R");
      $finish;
    end
  end
  // synthesis translate_on

  generate
  if ((WA == 32) && (SIGN == 1)) begin
    if (AREA_N_SPEED == 0) begin
      div_s_32_32 divs32 (
        .dividend(A),     // Bus [31 : 0]
        .divisor(B),      // Bus [31 : 0]
        .`QUOTIENT(quot),  // Bus [31 : 0]
        .rfd(rfd),
        .`REMAINDER(),
        .clk(CLK),
        .sclr(RESET),
        .ce(CE_IP_BLOCK));
    end else begin
      div_s_32_32_ii8 divs32 (
        .dividend(A),     // Bus [31 : 0]
        .divisor(B),      // Bus [31 : 0]
        .`QUOTIENT(quot),  // Bus [31 : 0]
        .rfd(rfd),
        .`REMAINDER(),
        .clk(CLK),
        .sclr(RESET),
        .ce(CE_IP_BLOCK));
    end
  end
  else if ((WA == 32) && (SIGN == 0)) begin
    if (AREA_N_SPEED == 0) begin
      div_u_32_32 divu32 (
        .dividend(A),     // Bus [31 : 0]
        .divisor(B),      // Bus [31 : 0]
        .`QUOTIENT(quot),  // Bus [31 : 0]
        .rfd(rfd),
        .`REMAINDER(),
        .clk(CLK),
        .sclr(RESET),
        .ce(CE_IP_BLOCK));
    end else begin
      div_u_32_32_ii8 divu32 (
        .dividend(A),     // Bus [31 : 0]
        .divisor(B),      // Bus [31 : 0]
        .`QUOTIENT(quot),  // Bus [31 : 0]
        .rfd(rfd),
        .`REMAINDER(),
        .clk(CLK),
        .sclr(RESET),
        .ce(CE_IP_BLOCK));
    end
  end
  else if ((WA == 16) && (SIGN == 1)) begin
    if (AREA_N_SPEED == 0) begin
      div_s_16_16 divs16 (
        .dividend(A),     // Bus [15 : 0]
        .divisor(B),      // Bus [15 : 0]
        .`QUOTIENT(quot),  // Bus [15 : 0]
        .rfd(rfd),
        .`REMAINDER(),
        .clk(CLK),
        .sclr(RESET),
        .ce(CE_IP_BLOCK));
    end else begin
      div_s_16_16_ii8 divs16 (
        .dividend(A),     // Bus [15 : 0]
        .divisor(B),      // Bus [15 : 0]
        .`QUOTIENT(quot),  // Bus [15 : 0]
        .rfd(rfd),
        .`REMAINDER(),
        .clk(CLK),
        .sclr(RESET),
        .ce(CE_IP_BLOCK));
    end
  end
  else if ((WA == 16) && (SIGN == 0)) begin
    if (AREA_N_SPEED == 0) begin
      div_u_16_16 divu16 (
        .dividend(A),     // Bus [15 : 0]
        .divisor(B),      // Bus [15 : 0]
        .`QUOTIENT(quot),  // Bus [15 : 0]
        .rfd(rfd),
        .`REMAINDER(),
        .clk(CLK),
        .sclr(RESET),
        .ce(CE_IP_BLOCK));
    end else begin
      div_u_16_16_ii8 divu16 (
        .dividend(A),     // Bus [15 : 0]
        .divisor(B),      // Bus [15 : 0]
        .`QUOTIENT(quot),  // Bus [15 : 0]
        .rfd(rfd),
        .`REMAINDER(),
        .clk(CLK),
        .sclr(RESET),
        .ce(CE_IP_BLOCK));
    end
  end
  else if ((WA == 8) && (SIGN == 1)) begin
    if (AREA_N_SPEED == 0) begin
      div_s_8_8 divs8 (
        .dividend(A),     // Bus [7 : 0]
        .divisor(B),      // Bus [7 : 0]
        .`QUOTIENT(quot),  // Bus [7 : 0]
        .rfd(rfd),
        .`REMAINDER(),
        .clk(CLK),
        .sclr(RESET),
        .ce(CE_IP_BLOCK));
    end else begin
      div_s_8_8_ii8 divs8 (
        .dividend(A),     // Bus [7 : 0]
        .divisor(B),      // Bus [7 : 0]
        .`QUOTIENT(quot),  // Bus [7 : 0]
        .rfd(rfd),
        .`REMAINDER(),
        .clk(CLK),
        .sclr(RESET),
        .ce(CE_IP_BLOCK));
    end
   end
   else if ((WA == 8) && (SIGN == 0)) begin
    if (AREA_N_SPEED == 0) begin
      div_u_8_8 divu8 (
        .dividend(A),     // Bus [7 : 0]
        .divisor(B),      // Bus [7 : 0]
        .`QUOTIENT(quot),  // Bus [7 : 0]
        .rfd(rfd),
        .`REMAINDER(),
        .clk(CLK),
        .sclr(RESET),
        .ce(CE_IP_BLOCK));
    end else begin
      div_u_8_8_ii8 divu8 (
        .dividend(A),     // Bus [7 : 0]
        .divisor(B),      // Bus [7 : 0]
        .`QUOTIENT(quot),  // Bus [7 : 0]
        .rfd(rfd),
        .`REMAINDER(),
        .clk(CLK),
        .sclr(RESET),
        .ce(CE_IP_BLOCK));
    end
  end
  endgenerate
  // shift in RESULT_READY signals based on START
  // thus these are aligned with the computation result
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      RESULT_READY_int <= 0;
    end else begin
      if (CE_IP_BLOCK)
        RESULT_READY_int <= (RESULT_READY_int << 1) | (START & START_ACK);
      else
        RESULT_READY_int <= RESULT_READY_int;
    end
  end

  // shift register instantiation
  shiftreg #(.WA(WR), .WR(WR), .DEPTH(DEPTH)) sr (
    .A(quot),
    .R(DATA_SHIFTREG_TO_QUEUE),
    .RESET(RESET),
    .CLK(CLK),
    .CE(CE_SHIFTREG),
    .START(RESULT_READY_int[LATENCY-1]),
    .RESULT_READY(RESULT_READY_SHIFTREG));

  `include "functions.v"

endmodule
