// *************************************************************
// Dieses Modul nimmt eine Flu"skontrolle zwischen hintereinander-
// geschalteten Eingabe- und Ausgabe-Streams vor.  Dabei werden aus
// den beiden Stall-Signalen das Enable des jeweils anderen Streams
// generiert.  Ein globales Enable steuert das gesamte Konstrukt: 
// Wenn es High ist, wird zur n"achsten Posedge der CLK ein Datum
// gelesen und gleichzeitig ein Datum geschrieben.  Das vom
// Eingabe-Strom gelesene Datum wird "uber den Port READ_DATA an das
// Modul "ubergeben (ggf.  lokal gepuffert).  Der Port WRITE_DATA kann
// dann an den Ausgabe-Stream angeschlossen werden.
// 
// Bei gepipelinter Ausf"uhrung (mehre Takte Latenz der Logik zwischen
// Eingabe- und Ausgabestream) m"ussen die READ_STALLs innerhalb der
// Pipeline mitgef"uhrt werden.  Sie halten so den Ausgabe-Stream an,
// bis tats"achlich zu schreibende Daten anstehen.  Das READ_ENABLE
// Signal mu"s aber sowohl an die Pipeline (Verhindern des Vorr"uckens
// bei FALSE) als auch an den Eingabe-Stream angeschlossen werden
// (Bremsen der urspr"unglichen Datenquelle).
// *************************************************************

`timescale 1ns/10ps

`include "marcdefs.v"

module flowcontrol

          (input                    CLK,
           input                    ENABLE,       // globales ENABLE
           input                    READ_STALL,   // Stall vom Readstream
           input                    WRITE_STALL,  // Stall vom Writestream
           input [`DATA_WIDTH-1:0]  READ_DATA,    // Daten vom Readstream
           output                   READ_ENABLE,  // Enable fr Readstream
           output                   WRITE_ENABLE, // Enable fr Writestream
           output [`DATA_WIDTH-1:0] WRITE_DATA);  // Daten fr Writestream
                      



  reg                               READ_STALL_REG;
  reg                               WRITE_STALL_REG;
  reg [`DATA_WIDTH-1:0]             READ_DATA_REG;
  reg                               ENABLE_REG;


        // Starte/Stoppe Read-Stream direkt mit globalem Enable
        // Halte ihn auch sofort bei Write-Stall an
  wire READ_ENABLE = ENABLE & ~WRITE_STALL;

        // Starte/Stoppe Write-Stream verz"ogert zu globalem Enable:
        //      -- im ersten Takt sind noch keine Daten zu schreiben
        //      -- im letzten Takt sind noch Daten zu schreiben (vor Stop)
        // Schreibe bei trotz Read-Stall noch die aktuellen Daten
        // und halte erst dann (einen Takt verz"ogert) an
  wire WRITE_ENABLE = ENABLE_REG & ~READ_STALL_REG;

  // Multiplexer fr Restart nach Write-Stall
        // Falls ein Write-Stall im letzten Takt aufgeten ist,
        // schreibe zun"achst die Daten, die vor dem Write-Stall
        // gelesen wurden, aber nicht mehr geschrieben werden konnten.
        // (Der Read-Stream liest ja beim Erkennen des Write-Stalls
        // schon das n"achste Datum ein, das vorherige mu"s also
        // gepuffert werden)
        // Im Normalbetrieb schreibe direkt die eingehenden Daten.
  wire [`DATA_WIDTH-1:0] WRITE_DATA = WRITE_STALL_REG 
                                        ? READ_DATA_REG
                                        : READ_DATA;

  // Datenpfad bei WRITE_STALL puffern
        // READ_DATA_REG merkt sich das letzte
        // vor dem Write-Stall vom Read-Stream
        // gelesenen Daten. Falls es zum Write-Stall kommt,
        // kann es nach dem Restart aus diesem Puffer geschrieben werden.
  always @(posedge CLK) begin
    if (~WRITE_STALL_REG) begin
      READ_DATA_REG <= READ_DATA;
    end
  end

  // Stallsignale verzgern
        // ... werden an verschiedenen Stellen (s.o.) gebraucht ...
  always @(posedge CLK) begin
    READ_STALL_REG <= READ_STALL;
    WRITE_STALL_REG <= WRITE_STALL;
  end

  // ENABLE verzgern
        // Falls kein Stall auftritt, verz"ogere das Enable nur,
        // anderenfalls halte es solange fest, bis alle Stalls
        // verschwunden sind. Auf diese Weise wird garantiert,
        // da"s alle Daten (trotz Stalls) noch geschrieben werden,
        // obwohl die User-Schaltung die Operation schon beendet
        // glaubt und das globale Enable wegnimmt.
  always @(posedge CLK) begin
    if (~READ_STALL & ~WRITE_STALL) begin
      ENABLE_REG <= ENABLE;
    end
  end

`ifdef synthesis
`else
  initial begin
    READ_STALL_REG <= 0;
    WRITE_STALL_REG <= 0;
    READ_DATA_REG <= 0;
    ENABLE_REG <= 0;
  end
`endif

endmodule

