// E.I.S. / ESA modlib for COMRADE and HardScale
  //
  //// Description:
  // ----------------
  // Module which just routes through data and tokens; it has no internal buffers.
  // A gate node always has exactly one data and one control predecessor and one
  // data successor. It is (e.g.) used to forward data to mux inputs.
  //
  // Version History:
  // ----------------
  // 100826: BT: intial release

`timescale 1ns / 1ns

module gate
         #(parameter
                      WA                   = 32,
                      WB                   = 1,  // NIN
                      NIN                  = WB,
                      WR                   = WA, // 1 for logic operations
                      PSEUDO_AT            = 0,
                      DEPTH                = 0,
                      QDEPTH               = 16,
                      TQDEPTH              = 16,
                      DATA_SPECULATION     = 0,
                      SPECULATION_BOUNDARY = 0,
                      STATIC_CT            = 0,
                      START_CTRL_IN        = 0,
                      SIGN                 = 0
          )(
           input  wire    [(WA*NIN-1):0] A,
           output wire    [(WR-1)    :0] R,
           input  wire                   RESET,
           input  wire                   CLK,
           input  wire                   CE,
           input  wire                   START,
           output wire                   START_ACK,
           input  wire                   START_CTRL,
           output wire                   START_CTRL_ACK,
           input  wire                   CANCEL,
           output reg                    CANCEL_ACK,
           output wire                   CANCEL_STATE,
           input  wire                   CANCEL_STATE_ACK,
           output wire                   CANCEL_STATE_CTRL,
           input  wire                   CANCEL_STATE_CTRL_ACK,
           output wire                   RESULT_READY,
           input  wire                   RESULT_READY_ACK);

  // define LATENCY of this module
  localparam LATENCY          = 1;
  // CANCEL BITS depending on LATENCY and shift register DEPTH
  localparam CANCEL_BITS      = min_1(ceil_log2(LATENCY+DEPTH-1));
  // QDEPTH_INT has to be at least 16
  localparam QDEPTH_INT       = min_16(QDEPTH);
  // set QBITS accordingly
  localparam QBITS            = (QDEPTH == 0) ? 0 : ceil_log2(QDEPTH_INT-1);
  // TQDEPTH_INT has to be at least 16
  localparam TQDEPTH_INT      = min_16(TQDEPTH);
  // set TQBITS accordingly
  localparam TQBITS           = (TQDEPTH == 0) ? 0 : ceil_log2(TQDEPTH_INT-1);
  // COMPUTATION_BITS depending  on LATENCY shift register DEPTH and queue QDEPTH
  localparam COMPUTATION_BITS = min_1(ceil_log2(LATENCY+DEPTH+QDEPTH_INT-1));

  
  wire  CONTROLLER_CANCEL = 1'b0;
  wire  START_CTRL_PSEUDO = 1'b0;

  `include "handshaking_assigns_verilog_only.v"

  // the operation
  wire [(WR-1):0] r_int = A;

  // example for complex logic operation
  //wire r_int = A[((WA*NIN)-1):(WA*(NIN-1))] & A[((WA*(NIN-1))-1):(WA*(NIN-2))];


  // shift register instantiation
  shiftreg #(.WA(WR), .WR(WR), .DEPTH(DEPTH)) sr (
    .A(r_int),
    .R(DATA_SHIFTREG_TO_QUEUE),
    .RESET(RESET),
    .CLK(CLK),
    .CE(CE_SHIFTREG),
    .START(START_SHIFTREG),
    .RESULT_READY(RESULT_READY_SHIFTREG));

  // queue instantiation
  disram_queue_non_spec #(.WR(WR), .QDEPTH(QDEPTH_INT), .QBITS(QBITS)) queue_out (
    .CLK(CLK),
    .RESET(RESET),
    .DIN(DATA_SHIFTREG_TO_QUEUE),
    .READ(data_read),
    .WRITE(data_write),
    .DOUT(DATA_QUEUE_TO_OUT),
    .EMPTY(data_empty),
    .FULL(data_full));


  `include "functions.v"

endmodule
