// E.I.S. / ESA modlib for COMRADE and HardScale
  //
  // Description:
  // ----------------
  // Handshaking for loop multiplexers.                                                                                                                                
  // CTs from ac_token_queue delete data before the data flows into the shiftreg.
  //
  // Version History:
  // ----------------
  // 100826: BT: intial release


  reg  init_uncommited;
   
  
  wire [(NIN-1):0] SELECTION;
  
  wire data_write;
  wire data_read;
  
  wire data_full;
  wire data_empty;
  wire RESULT_READY_SHIFTREG;
  wire increase_cancel_counter;
  wire decrease_cancel_counter;
  wire increase_computation_counter;
  wire decrease_computation_counter;

  wire [(WR-1)              :0]    DATA_QUEUE_TO_OUT;
  wire [(WR-1)              :0]    DATA_SHIFTREG_TO_QUEUE;
  reg  [(TQDEPTH_INT-1)     :0]    cancel_counter; 
  reg  [(COMPUTATION_BITS-1):0]    computation_counter;
  wire [(WR-1):0] r_int;
  integer i, j;

  reg INITIALIZABLE;
  
  wire space_for_result;
  wire data_read_conditional;
  
  wire ac_token_queue_in;
  wire ac_token_queue_read;
  wire ac_token_queue_write;
  wire ac_token_queue_out;
  wire ac_token_queue_empty;
  wire ac_token_queue_full;

        
  // a result is ready and consumed
  wire RESULT_CONSUMED = RESULT_READY & RESULT_READY_ACK;
  // a cancel request was accepted
  wire CANCEL_CONSUMED = CANCEL & CANCEL_ACK;
  // activate token in token queue
  wire ACTIVATE_TOKEN_AVAILABLE = ~ac_token_queue_out & ~ac_token_queue_empty;
  // cancel token in token queue
  wire CANCEL_TOKEN_AVAILABLE   =  ac_token_queue_out & ~ac_token_queue_empty;  
  wire CE_SHIFTREG = CE & (~data_full | data_read | (~RESULT_READY & (~RESULT_READY_SHIFTREG | ~ac_token_queue_empty)));
  // write new value to shiftreg, if init value comes in or if continue input comes in and AT is in queue 
  wire START_SHIFTREG = (START[0] & START_ACK[0]) | 
                        (START[1] & START_ACK[1] & ACTIVATE_TOKEN_AVAILABLE) ? 1'b1 : 1'b0;

// START_CTRL_IN must be 1
initial begin
  if (START_CTRL_IN != 1) begin
    $display("loop_mux: START_CTRL_IN is != 1: %d", START_CTRL_IN);
    $finish;
  end
end

// NIN and WB must be 2
initial begin
  if (NIN != 2 || WB != 2) begin
    $display("loop_mux: NIN (%d) and WB (%d) must both be 2", NIN, WB);
    $finish;
  end
end

// ONEHOT must be 1
initial begin
  if (ONEHOT != 1) begin
    $display("loop_mux: ONEHOT (%d) must be 1", ONEHOT);
    $finish;
  end
end
    
assign data_read_conditional = data_read;

// assign shift register output or queue output to module output (depending on output queue status)
assign R         = data_empty ? DATA_SHIFTREG_TO_QUEUE : DATA_QUEUE_TO_OUT;

assign increase_computation_counter = ((START[0] & START_ACK[0]) || (START[1] & START_ACK[1])) && // data is coming in, and
                                      ~CANCEL_TOKEN_AVAILABLE; // the incoming data is not deleted by a stored CT
assign decrease_computation_counter = data_read_conditional & (~data_empty | data_write); // data is forwarded to successor modules

// ----- INITIALIZABLE -----
  generate
  if (DATA_SPECULATION == 0) begin
    // ----- INITIALIZABLE -----
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        INITIALIZABLE   <= 1'b1;
        init_uncommited <= 1'b1;
      end else begin
        // deactivate INITIALIZABLE after data has been consumed from init input
        if (INITIALIZABLE & START[0] & START_ACK[0]) begin
          INITIALIZABLE <= 1'b0;
        end
        // reactivate INITIALIZABLE after loop-end CT has left via CANCEL_STATE or has been extinguished
        else if (~INITIALIZABLE &
                 ((CANCEL_STATE[1] & CANCEL_STATE_ACK[1]) |
                  (START[1] & START_ACK[1] & ~ac_token_queue_empty & ac_token_queue_out) |
                  (~ac_token_queue_empty & ac_token_queue_out & NO_CT))) begin
          INITIALIZABLE <= 1'b1;
        end else begin
          INITIALIZABLE <= INITIALIZABLE;
        end
      end
    end
  end  
  endgenerate
   

// ----- START_ACK -----
assign space_for_result = (DEPTH > 0)  ? ((~data_full & ((DATA_SPECULATION == 0))) | ((QDEPTH == 0) & ~RESULT_READY))
                        : (QDEPTH > 0) ? (~data_full & ((DATA_SPECULATION == 0)))
                        :                RESULT_READY & RESULT_READY_ACK;

assign START_ACK[0] = START[0] &        // accept data from init input, if
                      INITIALIZABLE &   // INITIALIZABLE is true, and
                      space_for_result; // the loop mux is able to hold the data

generate
if (STATIC_CT == 0) begin
  // dynCT
  assign START_ACK[1] = START[1] &                 // accept data from continue input, if
                        ~INITIALIZABLE &           // INITIALIZABLE is false, and
                        space_for_result &         // the loop mux is able to hold the data, and
                        ~ac_token_queue_empty & ~ac_token_queue_out; // there is an AT in the token queue (which is now consumed)
end else begin
  // statCT
  assign START_ACK[1] = START[1] &                 // accept data from continue input, if
                        ~INITIALIZABLE &           // INITIALIZABLE is false, and
                        space_for_result &         // the loop mux is able to hold the data, and
                        ~ac_token_queue_empty;                 // there is an AT or CT (!) in the token queue (which is now consumed,
                                                   //   if it is as AT, or extinguished, if it is a CT)
end
endgenerate

// ----- START_CTRL_ACK -----
assign START_CTRL_ACK = ~ac_token_queue_full & START_CTRL;
                                      
generate
  // ----- CANCEL_STATE -----
  if (STATIC_CT == 1) begin
    // static CTs
    // never set CANCEL_STATE
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        CANCEL_STATE <= 1'b0;
      end else begin
        CANCEL_STATE <= 1'b0;
      end
    end
  end else begin
    // dynamic CTs
    
    // never set CANCEL_STATE for init input
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        CANCEL_STATE[0] <= 1'b0;
      end else begin
        CANCEL_STATE[0] <= 1'b0;
      end
    end
    
    // continue input: move CT from ac_token_queue to CANCEL_STATE[1]
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        CANCEL_STATE[1] <= 1'b0;
      end else begin
        CANCEL_STATE[1] <= (~ac_token_queue_empty & ac_token_queue_out & ~NO_CT)           ? 1 // set CANCEL_STATE[1], if CT is removed from ith TQ,
                                                                             //   and NO_CT is not set;
                             : (CANCEL_STATE[1] & CANCEL_STATE_ACK[1]) ? 0 // reset CANCEL_STATE[1], if reset has been signalized
                             : CANCEL_STATE[1];
      end
    end
  end
endgenerate

  // ----- CANCEL_STATE_CTRL -----
  assign CANCEL_STATE_CTRL = 1'b0;
  
  // ----- RESULT_READY -----
  generate
  if (DATA_SPECULATION == 0 | QDEPTH != 0) begin
  assign RESULT_READY = data_empty ? RESULT_READY_SHIFTREG : 1'b1;
  end
  endgenerate
  
  // ----- read / write -----
  assign data_write = RESULT_READY_SHIFTREG;
  // read from output queue when a result is ready (may read from empty queue)
  
  generate
  if (DATA_SPECULATION == 0) begin 
  assign data_read = RESULT_CONSUMED;  
  end
  endgenerate
  
  // ----- increase / decrease -----
  assign decrease_cancel_counter = 1'b0;
  assign increase_cancel_counter = 1'b0;
  
  // ----- TOKENS_EMPTY -----
  assign TOKENS_EMPTY = 1'b0;
  
  // ----- CURRENT_TOKEN -----
  assign CURRENT_TOKEN = 1'b0;
  
  // ----- CURRENT_TOKEN_PSEUDO -----
  assign CURRENT_TOKEN_PSEUDO = 1'b0;

  // CANCEL_ACK
  assign CANCEL_ACK = ~ac_token_queue_full; // accept CTs as long as they can be stored in the ac_token_queue

  assign SELECTION = INITIALIZABLE & START[0] & START_ACK[0] ? 2'b01
                   : ~INITIALIZABLE  & START[1] & START_ACK[1] ? 2'b10
                   : 2'b00;

// queue to keep activate/cancel tokens
sr_queue #(.WR(1), .QDEPTH(16), .QBITS(4)) ac_token_queue ( 
  .CLK(CLK),         
  .RESET(RESET), 
  .DIN(ac_token_queue_in), 
  .READ(ac_token_queue_read), 
  .WRITE(ac_token_queue_write), 
  .DOUT(ac_token_queue_out),
  .EMPTY(ac_token_queue_empty), 
  .FULL(ac_token_queue_full));

  assign ac_token_queue_in = CANCEL_CONSUMED ? 1'b1   // entering CT
                     : START[1] & START_ACK[1] ? 1'b0 // entering AT
                     : 1'b0; // dummy
  
  assign ac_token_queue_read = (START[1] & START_ACK[1] & ~ac_token_queue_empty) |              // read AT (or CT), when incoming continue data is stored in shiftreg (or extinguished)
                               (~CANCEL_STATE[1] & CANCEL_TOKEN_AVAILABLE & (STATIC_CT == 0)) | // read CT, when CT flows into CANCEL_STATE for continue input
                               (~ac_token_queue_empty & ac_token_queue_out & (NO_CT == 1) & ~INITIALIZABLE); // read CT, when NO_CT==1 (at the same time, INITIALIZABLE is reset to 1)
  
  assign ac_token_queue_write = (START_CTRL & START_CTRL_ACK) | // write AT if AT for continue input comes in
                                CANCEL_CONSUMED;               // write CT if CT for continue input comes in
  
  // count ongoing computations  
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      computation_counter <=  {(COMPUTATION_BITS){1'b0}};
    end else begin
      case ({decrease_computation_counter, increase_computation_counter}) 
           2'b00: computation_counter <= computation_counter;
           2'b01: if (computation_counter != {(COMPUTATION_BITS){1'b1}})
                     computation_counter <= computation_counter + 1'b1;
           2'b10: if (computation_counter != {(COMPUTATION_BITS){1'b0}})
                     computation_counter <= computation_counter - 1'b1;              
           2'b11: computation_counter <= computation_counter; 
      endcase
    end
  end  
   
 
  // on CANCEL request the signal CANCEL_ACK is assertet on acceptance
  // acceptance is either the increase of an internal counter or the forwarding
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      cancel_counter <=  {(TQDEPTH_INT){1'b0}};
    end else if (data_empty) begin
      case ({decrease_cancel_counter, increase_cancel_counter}) 
           2'b00: begin
              cancel_counter <= cancel_counter;
           end
           2'b01: begin
              if (cancel_counter != {(TQDEPTH_INT){1'b1}}) begin  
                 cancel_counter <= cancel_counter + 1'b1;
              end
           end         
           2'b10: begin
              if (cancel_counter != {(TQDEPTH_INT){1'b0}}) begin
                 cancel_counter <= cancel_counter - 1'b1;
              end
           end         
           2'b11: begin
                cancel_counter <= cancel_counter;
           end
      endcase
    end
  end

  // queue instantiation    
  generate
  if (DATA_SPECULATION == 0) begin 
  disram_queue_non_spec #(.WR(WR), .QDEPTH(QDEPTH_INT), .QBITS(QBITS)) queue_out ( 
    .CLK(CLK),
    .RESET(RESET),
    .DIN(DATA_SHIFTREG_TO_QUEUE),
    .READ(data_read),
    .WRITE(data_write),
    .DOUT(DATA_QUEUE_TO_OUT),
    .EMPTY(data_empty),
    .FULL(data_full));
  end
  endgenerate
