// E.I.S. / ESA modlib for COMRADE and HardScale
  //
  // Description:
  // ----------------
  // Handshaking for multiplexers.
  //
  // Behavior:
  // - Data is read from input i if START[i] & START_ACK[i]; during the same cycle, the result is computed and stored in the shift register sr.
  //   (If sr has DEPTH 0, the next step follows in the same cycle.)
  // - From the sr, the result moves on to the output data queue (which also can have size 0, i.e. QDEPTH=0).
  // - If the output data queue contains or forwards a result AND there is an AT in the ac_token_queue, then
  //   RESULT_READY is set. As a special case, if DEPTH=0 and QDEPTH=0, RESULT_READY is already set if there is
  //   valid output data, not needing an AT from the token queue (this is needed to prevent deadlocks in combinatoric
  //   handshaking).
  // - If PSEUDO_AT=1, then CONTROLLER_CANCEL may not be 1'bx!
  // - Select input is buffered in sel_queue and at the same time AT/CT are buffered in token_queues (one per input);
  //   thus, CTs can move to predecessors independently of incoming valid data
  //
  // Version History:
  // ----------------
  // 100826: BT: intial release


  wire [NIN-1:0] sel_in;
  wire [NIN-1:0] sel_out;
  wire           sel_empty;
  wire           sel_full;
  wire           sel_read;
  wire           sel_write;

  wire [NIN-1:0] tq_in;
  wire [NIN-1:0] tq_out;
  wire [NIN-1:0] tq_empty;
  wire [NIN-1:0] tq_full;
  wire [NIN-1:0] tq_read;
  wire [NIN-1:0] tq_write;
  wire [NIN-1:0] tq_delete_at;
  wire [NIN-1:0] tq_delete_ct;

  wire RESULT_READY_SHIFTREG;            // data available at the shift register
  wire CE_SHIFTREG;
  wire space_for_result;
  wire data_read_conditional;
  
  // wire definitions output queue
  wire [(WR-1):0]   DATA_SHIFTREG_TO_QUEUE;
  wire [(WR-1):0]   DATA_QUEUE_TO_OUT;
  wire              data_read;
  wire              data_write;
  wire              data_empty;
  wire              data_full;
  
  wire increase_cancel_counter;
  wire decrease_cancel_counter;
  wire increase_computation_counter;
  wire decrease_computation_counter;
  

  reg  [(TQDEPTH_INT-1)     :0]  cancel_counter;
  reg  [(COMPUTATION_BITS-1):0]  computation_counter;
  reg  [(WR-1)              :0]  r_int;


  wire [(NIN-1)             :0]  SELECTION  = sel_empty ? sel_in : sel_out;
  
  // a result is ready and consumed
  wire RESULT_CONSUMED = RESULT_READY & RESULT_READY_ACK;

  // data available and data accept
  wire START_SHIFTREG = ~sel_empty & sel_read & (|sel_out);

  // assign shift register output or queue output to module output (depending on output queue status)
  assign R         = data_empty ? DATA_SHIFTREG_TO_QUEUE : DATA_QUEUE_TO_OUT;

  assign increase_computation_counter = START_SEL & START_SEL_ACK;
  assign decrease_computation_counter = data_read_conditional & (~data_empty | data_write);

  generate
  if (START_CTRL_IN == 0) begin
    // the shiftreg is allowed to shift out data when the global enable is set
    // and either the output queue can buffer the data
    // there will be free space due to a concurrent read
    // or RESULT_READY remains low due to a cancellation
    assign CE_SHIFTREG = CE & (~data_full | data_read | ~RESULT_READY);
  end

  if (START_CTRL_IN == 1) begin
    // the shiftreg is allowed to shift out data when the global enable is set
    // and either the output queue can buffer the data
    // or there will be free space due to a concurrent read
    // or RESULT_READY remains low due to a cancellation and the cancel token is available
    // or the shiftreg has not data at its output
    assign CE_SHIFTREG = CE & (~data_full | data_read | (~RESULT_READY & (~RESULT_READY_SHIFTREG)));
  end
  endgenerate


  generate
  if (DATA_SPECULATION == 0) begin
    assign data_read_conditional = data_read;
  end
  endgenerate


  // ----- space_for_result -----

  generate
  if (DEPTH == 0 & QDEPTH == 0) begin
    assign space_for_result = RESULT_READY & RESULT_READY_ACK;
  end
  if (DEPTH == 0 & QDEPTH > 0) begin
    assign space_for_result = ~data_full;
  end
  if (DEPTH > 0 & QDEPTH == 0) begin
    assign space_for_result = (~data_full | ~RESULT_READY);
  end
  if (DEPTH > 0 & QDEPTH > 0) begin
    assign space_for_result = ~data_full;
  end
  endgenerate



  // ----- START_ACK -----

  genvar i_gen;
  generate

  if (START_CTRL_IN == 0 ) begin
    // no control edge; WARNING: for correct token flow, there must be a control edge!!!
    assign START_ACK = START && space_for_result;
  end

  if (START_CTRL_IN == 1 & STATIC_CT == 0 & DATA_SPECULATION == 0) begin
      // dynamic CTs
      for (i_gen = 0; i_gen < NIN; i_gen = i_gen + 1) begin: i_gen_START_ACK
        // accept START[i_gen], if
        assign START_ACK[i_gen] = ~tq_empty[i_gen] &                // the associated TQ is not empty, and
                                  ~sel_empty &                      // select queue is not empty, and
                                  (START[i_gen] & sel_out[i_gen]) & // selected START bit is 1 (using sel_out as bitmask), and
                                  space_for_result &                // there is space to buffer the result, and
                                  tq_out[i_gen];                    // AT is at output of selected TQ
      end
  end

  if (START_CTRL_IN == 1 & STATIC_CT == 1 & DATA_SPECULATION == 0) begin
      // static CTs
      for (i_gen = 0; i_gen < NIN; i_gen = i_gen + 1) begin: i_gen_START_ACK
        // accept START[i_gen], if
        assign START_ACK[i_gen] = ~tq_empty[i_gen] &                 // the associated TQ is not empty, and
                                  ((~sel_empty &                     // select queue is not empty, and
                                    START[i_gen] & sel_out[i_gen] &  // selected START bit is 1 (using sel_out as bitmask), and
                                    space_for_result &               // there is space to buffer the result, and
                                    tq_out[i_gen]) |                 // AT is at output of selected TQ, or
                                   (~tq_out[i_gen] &                 // CT is at output of selected TQ, and
                                    ~NO_CT[i_gen] &                  // that CT is authorized to cancel an input, and
                                    START[i_gen]));                  // associated START bit is 1
      end
  end

  endgenerate

  // ----- START_CTRL_ACK -----
  assign START_CTRL_ACK = 1'b0;


  generate
  if (START_CTRL_IN == 0) begin
    // no sel queue required
    assign sel_in       = 1'b0;
    assign sel_out      = 1'b0;
    assign sel_write    = 1'b0;
    assign sel_read     = 1'b0;
    assign sel_empty    = 1'b1;
    assign sel_full     = 1'b1;
    assign tq_in        = 0;
    assign tq_out       = 0;
    assign tq_write     = 0;
    assign tq_read      = 0;
    assign tq_delete_at = 0;
    assign tq_delete_ct = 0;
    assign tq_empty     = -1;
    assign tq_full      = -1;
  end

  // ----- sel_queue with control signals, and token_queues (one for each input) -----
  if (START_CTRL_IN == 1) begin

    // accept sel when there is space left in the sel queue and in the token queues
    assign START_SEL_ACK = START_SEL && ~sel_full && ~(|tq_full);

    // select signal
    assign sel_in = B; // B[i] = 1 means AT, =0 means CT! (this differs from the ac_token_queue semantics!)

    // store select value in queue
    assign sel_write = START_SEL & START_SEL_ACK;
  end
  
  if (START_CTRL_IN == 1 & DATA_SPECULATION == 0) begin 
    // delete select value from sel queue when
    assign sel_read = ~sel_empty &&                   // select queue is not empty, and
                      ((START & START_ACK & sel_out & // selected START bit is 1 (using sel_out as bitmask),
                      tq_out) ||                      // AT is at output of selected TQ, or
                      ~(|sel_out));                   // there are only CTs generated by this select value
  end
  
  endgenerate
  
  
  generate
  if (STATIC_CT == 0) begin // dynamic CTs
    // token queues: one for each input
    for (i_gen = 0; i_gen < NIN; i_gen = i_gen + 1) begin: i_gen_tq_dyn

      assign tq_in[i_gen] = B[i_gen];

      // store token in queue when select value is stored in select queue
      assign tq_write[i_gen] = sel_write;

      // delete AT from token_queue when
      assign tq_delete_at[i_gen] = START_ACK[i_gen]; // input is accepted

      assign tq_delete_ct[i_gen] = ~tq_empty[i_gen] &       // tq is not empty, and
                                     ~tq_out[i_gen] &         // tq outputs a CT, and
                                     (~CANCEL_STATE[i_gen] || // CANCEL_STATE[i_gen] currently contains no CT
                                     NO_CT[i_gen]);           // tq[i_gen] does not create CTs

      assign tq_read[i_gen] = tq_delete_at[i_gen] | tq_delete_ct[i_gen];

      // token queue instantiation
      sr_queue #(.WR(1), .QDEPTH(TQDEPTH_INT), .QBITS(TQBITS)) TQ_i_gen_dyn (
        .CLK(CLK),
        .RESET(RESET),
        .DIN(tq_in[i_gen]),
        .READ(tq_read[i_gen]),
        .WRITE(tq_write[i_gen]),
        .DOUT(tq_out[i_gen]),
        .EMPTY(tq_empty[i_gen]),
        .FULL(tq_full[i_gen]));

    end
  end

  if (STATIC_CT == 1) begin  // static CTs
    // token queues: one for each input
    for (i_gen = 0; i_gen < NIN; i_gen = i_gen + 1) begin: i_gen_tq_stat

      assign tq_in[i_gen] = B[i_gen];

      // store token in queue when select value is stored in select queue
      assign tq_write[i_gen] = sel_write;

      // delete AT from token_queue when
      assign tq_delete_at[i_gen] = START_ACK[i_gen]; // input is accepted

      // delete CT from token_queue when
      assign tq_delete_ct[i_gen] = ~tq_empty[i_gen] &                    // tq is not empty, and
                                     ~tq_out[i_gen] &                      // tq outputs a CT, and
                                     ((START[i_gen] & START_ACK[i_gen]) || // new data is coming in, or
                                     NO_CT[i_gen]);                        // tq[i] does not create CTs

      assign tq_read[i_gen] = tq_delete_at[i_gen] | tq_delete_ct[i_gen];

      // token queue instantiation
      sr_queue #(.WR(1), .QDEPTH(TQDEPTH_INT), .QBITS(TQBITS)) TQ_i_gen_stat (
        .CLK(CLK),
        .RESET(RESET),
        .DIN(tq_in[i_gen]),
        .READ(tq_read[i_gen]),
        .WRITE(tq_write[i_gen]),
        .DOUT(tq_out[i_gen]),
        .EMPTY(tq_empty[i_gen]),
        .FULL(tq_full[i_gen]));

    end
  end
  endgenerate

  generate
  if (START_CTRL_IN == 1) begin

    // queue to store select values
    sr_queue #(.WR(NIN), .QDEPTH(TQDEPTH_INT), .QBITS(TQBITS)) sel_queue (
      .CLK(CLK),
      .RESET(RESET),
      .DIN(sel_in),
      .READ(sel_read),
      .WRITE(sel_write),
      .DOUT(sel_out),
      .EMPTY(sel_empty),
      .FULL(sel_full));

  end

  // queue instantiation
  if (DATA_SPECULATION == 0) begin
  disram_queue_non_spec #(.WR(WR), .QDEPTH(QDEPTH_INT), .QBITS(QBITS)) queue_out (
    .CLK(CLK),
    .RESET(RESET),
    .DIN(DATA_SHIFTREG_TO_QUEUE),
    .READ(data_read),
    .WRITE(data_write),
    .DOUT(DATA_QUEUE_TO_OUT),
    .EMPTY(data_empty),
    .FULL(data_full));
  end
  endgenerate



  // ----- CANCEL_STATE / CANCEL_STATE_CTRL -----

  assign CANCEL_STATE_CTRL = 1'b0;

  generate
  if (START_CTRL_IN == 0) begin
    // no cancel forwarding
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        CANCEL_STATE <= 0;
      end
      else begin
        CANCEL_STATE <= 0;
      end
    end
  end

  if (START_CTRL_IN == 1 & STATIC_CT == 1) begin
    // static CTs - never set CANCEL_STATE
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        CANCEL_STATE <= 0;
      end
      else begin
        CANCEL_STATE <= 0;
      end
    end
  end
  if (START_CTRL_IN == 1 & STATIC_CT == 0) begin
    // dynamic CTs - move CT from token_queues into CANCEL_STATEs
    for (i_gen = 0; i_gen < NIN; i_gen = i_gen + 1) begin: i_cs
      always @(posedge CLK, posedge RESET) begin
        if (RESET) begin
          CANCEL_STATE[i_gen] <= 1'b0;
        end
        else begin
          // set CANCEL_STATE[i_gen], if CT is removed from ith TQ, and NO_CT[i_gen] is not set;
          // reset CANCEL_STATE[i_gen], if reset has been signaled
          CANCEL_STATE[i_gen] <= (tq_delete_ct[i_gen] & ~NO_CT[i_gen]) ? 1
                                                                       : (CANCEL_STATE[i_gen] & CANCEL_STATE_ACK[i_gen]) ? 0
                                                                                                                           : CANCEL_STATE[i_gen];
        end
      end
    end
  end
  endgenerate



  // ----- RESULT_READY -----
  // this node is always a speculation boundary
  assign RESULT_READY = data_empty ? RESULT_READY_SHIFTREG : 1'b1;

  // ----- read / write -----
  assign data_write = RESULT_READY_SHIFTREG;
  // read from output queue when a result is ready (may read from empty queue)

  generate
  if (DATA_SPECULATION == 0) begin
    assign data_read = RESULT_CONSUMED;
  end
  endgenerate

  // ----- increase / decrease -----
  // counter for debug and simulation purpose only: count SEL values is sel queue
  generate
  if (START_CTRL_IN == 1) begin
    assign decrease_cancel_counter = sel_read;
    assign increase_cancel_counter = START_SEL & START_SEL_ACK;
  end else begin
    assign decrease_cancel_counter = 1'b0;
    assign increase_cancel_counter = 1'b0;
  end
  endgenerate


  // ----- TOKENS_EMPTY -----
  assign TOKENS_EMPTY = 1'b0;

  // ----- CURRENT_TOKEN -----
  assign CURRENT_TOKEN = 1'b0;

  // ----- CURRENT_TOKEN_PSEUDO -----
  assign CURRENT_TOKEN_PSEUDO = 1'b0;
  // CANCEL_ACK
  assign CANCEL_ACK = 1'b0;



  // ----- CANCEL COUNTER -----

  // the cancel counter takes care of the pending cancel requests
  // on CANCEL request the signal CANCEL_ACK is assertet on acceptance
  // acceptance is either the increase of an internal counter or the forwarding
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      cancel_counter <=  {(TQDEPTH_INT){1'b0}};
    end else if (data_empty) begin
      case ({decrease_cancel_counter, increase_cancel_counter})
           2'b00: begin
              cancel_counter <= cancel_counter;
           end
           2'b01: begin
              if (cancel_counter != {(TQDEPTH_INT){1'b1}}) begin
                 cancel_counter <= cancel_counter + 1'b1;
              end
           end
           2'b10: begin
              if (cancel_counter != {(TQDEPTH_INT){1'b0}}) begin
                 cancel_counter <= cancel_counter - 1'b1;
              end
           end
           2'b11: begin
                cancel_counter <= cancel_counter;
           end
      endcase
    end
  end


  // ----- COMPUTATION COUNTER -----

  // the computation counter is required to determine whether a cancel needs to be forwared
  // via the cancel state output (dynamic cancel tokens only)
  // the coputation counter is also required in a special COMRADE case when there is a incomming control edge
  // since the data and memory dependencies both affect the START input CANCEL_STATE is used for the purpose to
  // allow the data to enter the module even if memory dependencies are unfulfilled
  generate
  if (START_CTRL_IN == 1 | STATIC_CT == 0) begin
    // count ongoing computations
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        computation_counter <=  {(COMPUTATION_BITS){1'b0}};
      end
      else begin
        case ({decrease_computation_counter, increase_computation_counter})
             2'b00: computation_counter <= computation_counter;
             2'b01: if (computation_counter != {(COMPUTATION_BITS){1'b1}})
                       computation_counter <= computation_counter + 1'b1;
             2'b10: if (computation_counter != {(COMPUTATION_BITS){1'b0}})
                       computation_counter <= computation_counter - 1'b1;
             2'b11: computation_counter <= computation_counter;
        endcase
      end
    end
  end
  else begin
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        computation_counter <=  {(COMPUTATION_BITS){1'b0}};
      end else begin
        computation_counter <= computation_counter;
      end
    end
  end
  endgenerate
  