// E.I.S. / ESA modlib for COMRADE and HardScale
  //
  // Description:
  // ----------------
  // Universal handshaking for verilog-only modules (i.e. no instanciated netlists).
  // ac_token_queue_in/out == 0 means AT; ac_token_queue_in/out == 1 means CT.
  //
  // Behavior:
  // - Data is read if START & START_ACK; during the same cycle, the result is computed and stored in the shift register sr.
  //   (If sr has DEPTH 0, the next step follows in the same cycle.)
  // - From the sr, the result moves on to the output data queue (which also can have size 0, i.e. QDEPTH=0).
  // - If the output data queue contains or forwards a result AND there is an AT in the ac_token_queue, then
  //   RESULT_READY is set. As a special case, if DEPTH=0 and QDEPTH=0, RESULT_READY is already set if there is
  //   valid output data, not needing an AT from the token queue (this is needed to prevent deadlocks in combinatoric
  //   handshaking).
  // - If PSEUDO_AT=1, then CONTROLLER_CANCEL may not be 1'bx!
  //
  // Version History:
  // ----------------
  // 100826: BT: intial release


  wire data_read_conditional;
  wire pseudo_data_commit;


  // wire definitions token queue
  wire ac_token_queue_in;
  wire ac_token_queue_out;
  wire ac_token_queue_read;
  wire ac_token_queue_write;
  wire ac_token_queue_empty;
  wire ac_token_queue_full;

  // wire definition pseudo token queue
  wire pseudo_token_queue_in;
  wire pseudo_token_queue_out;
  wire pseudo_token_queue_empty;
  wire pseudo_token_queue_full;


  // wire definitions output queue
  wire [(WR-1):0]   DATA_SHIFTREG_TO_QUEUE;
  wire [(WR-1):0]   DATA_QUEUE_TO_OUT;
  wire              data_read;
  wire              data_write;
  wire              data_empty;
  wire              data_full;

  // cancel counter
  wire increase_cancel_counter;
  wire decrease_cancel_counter;

  // internal signals
  wire RESULT_READY_SHIFTREG;            // data available at the shift register
  wire CE_SHIFTREG;
  wire START_SHIFTREG;
  


  // registers
  reg  [(CANCEL_BITS-1)     :0]    cancel_counter;             // counts outstanding cancel operations
  reg  [(COMPUTATION_BITS-1):0]    computation_counter;        // counts number of ongoing computation
  reg                              cancel_state_keep;          // keep cancel state high until acknowledged
  reg                              cancel_state_ctrl_keep;     // keep cancel state control high until acknowledged
  reg                              cancel_to_process;          // there is a cancel that has not been fully processed


  // SIMPLIFICATION DEFINITIONS FOR BETTER READABILITY

  // a result is ready and consumed
  wire RESULT_CONSUMED = RESULT_READY & RESULT_READY_ACK;
  // a cancel request was accepted
  wire CANCEL_CONSUMED = CANCEL & CANCEL_ACK;
  // activate token in token queue
  wire ACTIVATE_TOKEN_AVAILABLE = ~ac_token_queue_out & ~ac_token_queue_empty;
  // cancel token in token queue
  wire CANCEL_TOKEN_AVAILABLE   =  ac_token_queue_out & ~ac_token_queue_empty;
  // pseudo token in token queue
  wire PSEUDO_TOKEN_AVAILABLE  = pseudo_token_queue_out & ~ac_token_queue_empty;
  // the modlib supports immediate data forwarding on activate token arrival
  wire ACTIVATE_TOKEN_ARRIVAL   = START_CTRL & START_CTRL_ACK & ac_token_queue_empty;
  // this behaviour has not been implemented for cancel tokens
  //wire CANCEL_token_queue_ENTERING  = CANCEL_CONSUMED & ac_token_queue_empty;

  // when data enters the module the computation counter increases
  wire increase_computation_counter = START & START_ACK;
  // decrease computation count when data is removed from the data queue
  // or data from the output of the shift register is removed (due to a cancel)
  wire decrease_computation_counter = (data_read_conditional & (~data_empty | data_write)) | (RESULT_READY_SHIFTREG & ~data_write);


  // OUTPUT ASSIGNMENTS

  // assign shift register output or queue output to module output (depending on output queue status)
  assign R             = data_empty ? DATA_SHIFTREG_TO_QUEUE : DATA_QUEUE_TO_OUT;


  // INTERNAL ASSIGNMENTS

  generate
  if (START_CTRL_IN == 0) begin
    // the shiftreg is allowed to shift out data when the global enable is set
    // and either the output queue can buffer the data
    // there will be free space due to a concurrent read
    // or RESULT_READY remains low due to a cancellation
    assign CE_SHIFTREG = CE & (~data_full | data_read | ~RESULT_READY);
  end

  if (START_CTRL_IN == 1) begin
    // the shiftreg is allowed to shift out data when the global enable is set
    // and either the output queue can buffer the data
    // or there will be free space due to a concurrent read
    // or RESULT_READY remains low due to a cancellation and the cancel token is available
    // or the shiftreg has not data at its output
    assign CE_SHIFTREG = CE & (~data_full | data_read | (~RESULT_READY & (~RESULT_READY_SHIFTREG | ~ac_token_queue_empty)));
  end
  endgenerate

  generate
  if (DEPTH > 0) begin
    // data available and data accept
    assign START_SHIFTREG = START & START_ACK;
  end
  if (DEPTH == 0) begin
    // ensure passtrough when the shift register is disabled
    assign START_SHIFTREG = START;
  end
  endgenerate


  generate
  if (DATA_SPECULATION == 0 | PSEUDO_AT == 0) begin
    assign pseudo_data_commit = 1'b0;
  end
  endgenerate


  generate
  if (DATA_SPECULATION == 0) begin
    assign data_read_conditional    = data_read;
  end
  endgenerate

  generate
  if (PSEUDO_AT == 0) begin
    assign pseudo_token_queue_out = 1'b0;
  end
  endgenerate

  // ----- AC_TOKEN_QUEUE AND CONTROL SIGNALS -----
  generate

  if (START_CTRL_IN == 1) begin
    // activate/cancel queue required, set when there is space left in the token queue
    assign  START_CTRL_ACK = START_CTRL & ~ac_token_queue_full;
    // ac_token_queue_in == 1 means CANCEL, ac_token_queue_in == 0 means ACTIVATE
    assign  ac_token_queue_in = CANCEL_CONSUMED;
    // ac_token_queue_write validates ac_token_queue_in on token arrival
    assign  ac_token_queue_write  = CANCEL_CONSUMED || (START_CTRL & START_CTRL_ACK);
  end

  if (START_CTRL_IN == 1 & PSEUDO_AT == 0) begin
    // no pseudo tokens
    assign  pseudo_token_queue_in = CONTROLLER_CANCEL;
  end

  if (START_CTRL_IN == 1 & PSEUDO_AT == 1) begin
    // pseudo_token_queue_in == 1 means pseudoAT, pseudo_token_queue_in == 0 means non-pseudo AT or CT
    // second meaning: the module is cancelled by a cancelled controller
    assign  pseudo_token_queue_in = START_CTRL_PSEUDO | CONTROLLER_CANCEL;
  end

  // dynamic CTs:
  if (START_CTRL_IN == 1 & STATIC_CT == 0 & PSEUDO_AT == 0) begin
    // delete token when corresponding data is read or cancelled from the data output queue
    assign  ac_token_queue_read = data_read |
                       // or directly from the sr, or
                       (RESULT_READY_SHIFTREG & ~data_write) |
                       // a cancel token from the ac_token_queue is consumed by the predecessors
                       (CANCEL_STATE & CANCEL_STATE_ACK);
  end

  if (START_CTRL_IN == 1 & STATIC_CT == 0 & PSEUDO_AT == 1) begin
    // delete token when corresponding data is read (or cancelled) from the data output queue
    assign  ac_token_queue_read = data_read |
                       // or directly from the shiftreg
                       (RESULT_READY_SHIFTREG & ~data_write & (CANCEL_TOKEN_AVAILABLE | (ACTIVATE_TOKEN_AVAILABLE & ~PSEUDO_TOKEN_AVAILABLE))) |
                       // or a pseudo AT creates dummy data which is consumed by the successors
                       (RESULT_CONSUMED & PSEUDO_TOKEN_AVAILABLE) |
                       // or a cancel token from the ac_token_queue is consumed by the predecessors
                       (CANCEL_STATE & CANCEL_STATE_ACK);
  end

  // static CTs: not necessary to delete a cancel token (no passing to its predecessor)
  if (START_CTRL_IN == 1 & STATIC_CT == 1 & PSEUDO_AT == 0) begin
    // delete token when corresponding data is read from the data output queue or directly from the shiftreg
    assign  ac_token_queue_read = data_read | (RESULT_READY_SHIFTREG & ~data_write);
  end

  if (START_CTRL_IN == 1 & STATIC_CT == 1 & PSEUDO_AT == 1) begin
    // delete token when corresponding data is read/cancelled from the data output queue or
    assign  ac_token_queue_read = data_read |
                       // or directly from the sr (no write to buffer, CT or pseudo AT)
                       (RESULT_READY_SHIFTREG & ~data_write & (CANCEL_TOKEN_AVAILABLE | (ACTIVATE_TOKEN_AVAILABLE & ~PSEUDO_TOKEN_AVAILABLE))) |
                       // or a pseudo AT creates dummy data which is consumed by the successors
                       (RESULT_CONSUMED & (ACTIVATE_TOKEN_AVAILABLE & pseudo_token_queue_out));
  end

  if (START_CTRL_IN == 0) begin
    // no token queue handling necessary
    assign  ac_token_queue_in            = 1'b0;
    assign  ac_token_queue_write         = 1'b0;
    assign  ac_token_queue_read          = 1'b0;
    assign  START_CTRL_ACK               = 1'b1;
    assign  pseudo_token_queue_in        = 1'b0;
  end

  endgenerate


  // ----- START_ACK -----
  generate

  if (START_CTRL_IN == 0 & STATIC_CT == 0 & DEPTH > 0) begin
    // the output queue is not full or if QDEPTH == 0 the shift register output is thrown away or canceled and no cancel forwarding is initiated
    assign START_ACK = START & (~data_full | ~RESULT_READY) & ~(CANCEL & data_empty & (computation_counter == {(COMPUTATION_BITS){1'b0}}));
  end


  if (STATIC_CT == 0 & DEPTH == 0 & QDEPTH == 0) begin
    // no output buffer available => accept data as soon as successor accepts data (to chain operators)
    assign START_ACK = START & RESULT_CONSUMED;
  end

  if (STATIC_CT == 0  & DEPTH == 0 & QDEPTH > 0) begin
    // accept data when the output queue is not full
    // prevent to signal RFD when a CANCEL token is currently forwarded
    assign START_ACK = START & ~data_full & ~(CANCEL & data_empty & (computation_counter == {(COMPUTATION_BITS){1'b0}}));  // !!! changes
  end

  if (START_CTRL_IN == 0 & STATIC_CT == 1 & DEPTH == 0 & QDEPTH == 0) begin
    // no cancels, no buffers and no latency - accept data as soon as previous data leaves the module
    assign START_ACK = START & RESULT_CONSUMED;
  end

  if (STATIC_CT == 1 & DEPTH == 0 & QDEPTH > 0) begin
    // accept data when the output queue is not full or data leaves the module currently
    assign START_ACK = START & (~data_full | RESULT_CONSUMED);  // !!! changes
  end

  if (START_CTRL_IN == 0 & STATIC_CT == 1 & DEPTH > 0) begin
    // accept data when the output queue is not full
    // or if QDEPTH == 0 the shift register output is thrown away or data leaves the module currently
    assign START_ACK = START & (~data_full | ~RESULT_READY | RESULT_CONSUMED); // !!! changes
  end


  if (START_CTRL_IN == 1 & STATIC_CT == 0 & DEPTH > 0) begin
    // accept data when the output queue is not full
    // or when invalid computation result at shiftreg's output is thrown away
    // prevent to signal RFD when a CANCEL token is currently forwarded or a CANCEL with immediate effect
    assign START_ACK = START & (~data_full | (~RESULT_READY & (~RESULT_READY_SHIFTREG | ~ac_token_queue_empty))) & ~(CANCEL & data_empty & (computation_counter == {(COMPUTATION_BITS){1'b0}}));
  end

  if (START_CTRL_IN == 1 & STATIC_CT == 1 & DEPTH == 0 & QDEPTH == 0) begin
    // no cancels, no buffers and no latency - accept data as soon as previous data leaves the module
    // or when a cancel token is awaiting data for elimination
    assign START_ACK = START & (RESULT_CONSUMED | CANCEL_TOKEN_AVAILABLE);
  end

  if (START_CTRL_IN == 1 & STATIC_CT == 1 & DEPTH > 0) begin
    // accept data when the output queue is not full
    // or if QDEPTH == 0 the shift register output is thrown away or data leaves the module currently
    assign START_ACK = START & (~data_full | (~RESULT_READY & (~RESULT_READY_SHIFTREG | ~ac_token_queue_empty)) | RESULT_CONSUMED);
  end

  endgenerate


  // ----- CANCEL_STATE / CANCEL_STATE_CTRL -----
  generate

  if (START_CTRL_IN == 0 & STATIC_CT == 0) begin
    // forward cancel when there is no running computation within this module
    assign CANCEL_STATE    = ((cancel_counter != {(CANCEL_BITS){1'b0}}) & (computation_counter == {(COMPUTATION_BITS){1'b0}}));
    // no cancel_state_ctrl notification in the without control edge
    assign CANCEL_STATE_CTRL = 1'b0;
  end

  if (START_CTRL_IN == 0 & STATIC_CT == 1) begin
    // no cancel forwarding in the static case
    assign CANCEL_STATE    = 1'b0;
    // no cancel_state_ctrl notification in the without control edge
    assign CANCEL_STATE_CTRL = 1'b0;
  end

  // dynamic CTs
  if (START_CTRL_IN == 1 & STATIC_CT == 0 & DATA_SPECULATION == 0) begin
    // provide CANCEL_STATE when the next token is a CANCEL TOKEN or no CANCEL_STATE_ACK obtained yet
    // possible since the cancel token is removed from the token queue on CANCEL_STATE_ACK
    assign CANCEL_STATE      = ((CANCEL_TOKEN_AVAILABLE & ~cancel_to_process) | cancel_state_keep) & computation_counter == {(COMPUTATION_BITS){1'b0}};
    // provide CANCEL_STATE_CTRL when the next token is a CANCEL TOKEN or no CANCEL_STATE_CTRL_ACK obtained yet
    // keep cancel state ctrl signal when cancel state was acknowledged first but avoid the sending the same token twice
    assign CANCEL_STATE_CTRL = (CANCEL_TOKEN_AVAILABLE & ~cancel_to_process) | cancel_state_ctrl_keep;
  end

  // static CTs
  if (START_CTRL_IN == 1 & STATIC_CT == 1 & DATA_SPECULATION == 0) begin
    // provide CANCEL_STATE when the next token is a CANCEL TOKEN or no CANCEL_STATE_ACK obtained yet
    // token is removed from queue on CANCEL_STATE_ACK
    assign CANCEL_STATE  = 1'b0;
    // provide CANCEL_STATE_CTRL when the next token is a CANCEL TOKEN or no CANCEL_STATE_CTRL_ACK obtained yet
    // avoid the sending the same token twice
    assign CANCEL_STATE_CTRL = (CANCEL_TOKEN_AVAILABLE & ~cancel_to_process) | cancel_state_ctrl_keep;
  end
  endgenerate


  // ----- RESULT_READY -----
  generate

  // no incoming control edge
  if (START_CTRL_IN == 0 & STATIC_CT == 0 & SPECULATION_BOUNDARY == 0) begin
    // suppress RESULT_READY signal until all CANCEL signals processed
    // signal RESULT_READY when there is any queue content or computed data at the shift register
    assign RESULT_READY = (cancel_counter != {(CANCEL_BITS){1'b0}}) ? 1'b0 : data_empty ? RESULT_READY_SHIFTREG : 1'b1;
  end
  
  // no incoming control edge
  if (START_CTRL_IN == 0 & STATIC_CT == 0 & SPECULATION_BOUNDARY == 1) begin
    // suppress RESULT_READY signal until all CANCEL signals processed
    // signal RESULT_READY when there is any queue content or computed data at the shift register
    assign RESULT_READY = (cancel_counter != {(CANCEL_BITS){1'b0}}) ? 1'b0 : data_empty ? RESULT_READY_SHIFTREG : 1'b1;
  end

  // no incoming control edge / static ct -> no cancellation
  if (START_CTRL_IN == 0 & STATIC_CT == 1 & SPECULATION_BOUNDARY == 0) begin
    // signal RESULT_READY when there is any queue content or computed data
    assign RESULT_READY = data_empty ? RESULT_READY_SHIFTREG : 1'b1;
  end

  // no incoming control edge / static ct -> no cancellation
  if (START_CTRL_IN == 0 & STATIC_CT == 1 & SPECULATION_BOUNDARY == 1)  begin
    // signal RESULT_READY when there is any queue content or computed data
    assign RESULT_READY = data_empty ? RESULT_READY_SHIFTREG : 1'b1;
  end

  if (START_CTRL_IN == 1 & PSEUDO_AT == 0 & DEPTH == 0 & SPECULATION_BOUNDARY == 0) begin
    // signal RESULT_READY when there is an activate token in the queue or entering the module
    // otherwise signal RESULT_READY when there is any queue content or computed data flagged with an activate token
    assign RESULT_READY = data_empty ? RESULT_READY_SHIFTREG & (ACTIVATE_TOKEN_ARRIVAL | ACTIVATE_TOKEN_AVAILABLE)
                                     : ACTIVATE_TOKEN_ARRIVAL | ACTIVATE_TOKEN_AVAILABLE;
  end

  if (START_CTRL_IN == 1 & PSEUDO_AT == 0 & DEPTH == 0 & SPECULATION_BOUNDARY == 1) begin
    // suppress RESULT_READY signal until a non speculative activate token is available
    // signal RESULT_READY when there is an activate token in the queue or entering the module
    // otherwise signal RESULT_READY when there is any queue content or computed data flagged with an activate token
    assign RESULT_READY = data_empty ? (RESULT_READY_SHIFTREG & (ACTIVATE_TOKEN_ARRIVAL | ACTIVATE_TOKEN_AVAILABLE))
                                     : (ACTIVATE_TOKEN_ARRIVAL | ACTIVATE_TOKEN_AVAILABLE);
  end

  if (START_CTRL_IN == 1 & PSEUDO_AT == 0 & DEPTH != 0 & SPECULATION_BOUNDARY == 0) begin
    // data output queue is empty => RESULT_READY if the shift register contains/forwards a valid result and there is an AT in the token queue
    // data output queue contains some data => RESULT_READY if there is an AT in the token queue
    assign RESULT_READY = data_empty ? RESULT_READY_SHIFTREG & ACTIVATE_TOKEN_AVAILABLE
                                     : ACTIVATE_TOKEN_AVAILABLE;
  end

  if (START_CTRL_IN == 1 & PSEUDO_AT == 0 & DEPTH != 0 & SPECULATION_BOUNDARY == 1) begin
    // suppress RESULT_READY signal until a non speculative activate token is available
    // data output queue is empty= > RESULT_READY if the shift register contains/forwards a valid resultand there is an AT in the token queue
    // data output queue contains some data => RESULT_READY if there is an AT in the token queue
    assign RESULT_READY = data_empty ? (RESULT_READY_SHIFTREG & ACTIVATE_TOKEN_AVAILABLE)
                                     : ACTIVATE_TOKEN_AVAILABLE;
  end

  // incoming control edge
  if (START_CTRL_IN == 1 & PSEUDO_AT == 1 & DEPTH == 0 & SPECULATION_BOUNDARY == 0) begin
    // data output queue is empty => RESULT_READY if the shift register contains/forwards a valid result and there is an AT  or pseudo AT in the token queue
    // data output queue contains some data => RESULT_READY if there is an AT or a pseudo AT in the token queue
    assign RESULT_READY = data_empty ? (RESULT_READY_SHIFTREG & (ACTIVATE_TOKEN_ARRIVAL | ACTIVATE_TOKEN_AVAILABLE)) | (pseudo_token_queue_out & ACTIVATE_TOKEN_AVAILABLE)
                                     : ACTIVATE_TOKEN_ARRIVAL | ACTIVATE_TOKEN_AVAILABLE;
  end
  
  // incoming control edge
  if (START_CTRL_IN == 1 & PSEUDO_AT == 1 & DEPTH == 0 & SPECULATION_BOUNDARY == 1) begin
    // data output queue is empty => RESULT_READY if the shift register contains/forwards a valid result and there is an AT  or pseudo AT in the token queue
    // data output queue contains some data => RESULT_READY if there is an AT or a pseudo AT in the token queue
    assign RESULT_READY = data_empty ? ((RESULT_READY_SHIFTREG & (ACTIVATE_TOKEN_ARRIVAL | ACTIVATE_TOKEN_AVAILABLE)) | (pseudo_token_queue_out & ACTIVATE_TOKEN_AVAILABLE))
                                     : (ACTIVATE_TOKEN_ARRIVAL | ACTIVATE_TOKEN_AVAILABLE);
  end

  if (START_CTRL_IN == 1 & PSEUDO_AT == 1 & DEPTH != 0 & SPECULATION_BOUNDARY == 0) begin
    // suppress RESULT_READY signal when there is a cancel in the token queue
    // data output queue is empty => RESULT_READY if the shift register contains/forwards a valid result and there is an AT  or pseudo AT in the token queue
    // data output queue contains some data => RESULT_READY if there is an AT or a pseudo AT in the token queue
    assign RESULT_READY = data_empty ? ((RESULT_READY_SHIFTREG & ACTIVATE_TOKEN_AVAILABLE) | (pseudo_token_queue_out & ACTIVATE_TOKEN_AVAILABLE))
                                     : (ACTIVATE_TOKEN_ARRIVAL | ACTIVATE_TOKEN_AVAILABLE);
  end

  if (START_CTRL_IN == 1 & PSEUDO_AT == 1 & DEPTH != 0 & SPECULATION_BOUNDARY == 1) begin
    // suppress RESULT_READY signal when there is a cancel in the token queue
    // data output queue is empty => RESULT_READY if the shift register contains/forwards a valid result and there is an AT  or pseudo AT in the token queue
    // data output queue contains some data => RESULT_READY if there is an AT or a pseudo AT in the token queue
    assign RESULT_READY = data_empty ? ((RESULT_READY_SHIFTREG & ACTIVATE_TOKEN_AVAILABLE) | (pseudo_token_queue_out & ACTIVATE_TOKEN_AVAILABLE))
                                     : (ACTIVATE_TOKEN_ARRIVAL | ACTIVATE_TOKEN_AVAILABLE);
  end

  endgenerate


  // ----- read / write -----

  generate

  if (START_CTRL_IN == 0 & STATIC_CT == 0) begin
    // write to queue when there is no cancel avoiding this
    assign data_write = (cancel_counter == {(CANCEL_BITS){1'b0}}) ? RESULT_READY_SHIFTREG : 1'b0;

    // read next element when either the result has been consumed
    // or an element is canceled
    assign data_read = RESULT_CONSUMED | CANCEL_CONSUMED;
  end

  if (START_CTRL_IN == 0 & STATIC_CT == 1) begin
    // no cancel possible in this case (static cancel token without control edge)
    assign data_write = RESULT_READY_SHIFTREG;
    // read from output queue when a result is ready (may read from empty queue)
    assign data_read = RESULT_CONSUMED;
  end

  if (START_CTRL_IN == 1) begin
    // write to queue when there is no cancel avoiding this
    // no immediate cancellation
    assign data_write = (data_empty & CANCEL_TOKEN_AVAILABLE) ? 1'b0 : RESULT_READY_SHIFTREG;
  end

  if (START_CTRL_IN == 1 & PSEUDO_AT == 0) begin
    // read next element from the data output queue when either a result has been consumed, or an element is canceled due to a CT in the ac_token_queue
    assign data_read = RESULT_CONSUMED | (~data_empty & CANCEL_TOKEN_AVAILABLE);
  end

  if (START_CTRL_IN == 1 & PSEUDO_AT == 1) begin
    // either a (non-pseudo-AT) result has been consumed, or an element is canceled due to a CT in the ac_token_queue
    assign data_read = (RESULT_CONSUMED & ACTIVATE_TOKEN_AVAILABLE & ~pseudo_token_queue_out) |
                       (~data_empty & CANCEL_TOKEN_AVAILABLE);
  end

  endgenerate


  // ----- CANCEL COUNTER CONTROL -----

  generate
  // on CANCEL request the signal CANCEL_ACK is assertet on acceptance
  // acceptance is either the increase of an internal counter or the forwarding
  // of the cancel token via cancel_state / cancel_state_ctrl
  if (START_CTRL_IN == 0 & STATIC_CT == 0) begin
    // CANCEL received, but no entry in the output queue to cancel
    assign increase_cancel_counter      = (CANCEL_CONSUMED & data_empty);
    // supressed write to output queue or cancel forwarded successfully
    assign decrease_cancel_counter      = (RESULT_READY_SHIFTREG & ~data_write) | (CANCEL_STATE & CANCEL_STATE_ACK);
  end

  if (START_CTRL_IN == 0 & STATIC_CT == 1) begin
    // no cancel possible in this case
    assign increase_cancel_counter = 1'b0;
    assign decrease_cancel_counter = 1'b0;
  end

  // counter for debug and simulation purpose only
  if (START_CTRL_IN == 1) begin
    assign decrease_cancel_counter = ac_token_queue_read & CANCEL_TOKEN_AVAILABLE;
    assign increase_cancel_counter = CANCEL_CONSUMED;
  end
endgenerate


  // ----- TOKENS_EMPTY -----
  generate
  if (START_CTRL_IN == 0) begin
    assign TOKENS_EMPTY = 1'b1;
  end

  if (START_CTRL_IN == 1) begin
    assign TOKENS_EMPTY = ac_token_queue_empty;
  end
  endgenerate

  // ----- CURRENT_TOKEN -----
  generate
  if (START_CTRL_IN ==0) begin
    // no token queue => no information about stored tokens
    assign CURRENT_TOKEN = 1'b0;
  end

  if (START_CTRL_IN == 1) begin
    // ac_token_queue_out means CT, ~ac_token_queue_out means AT => CURRENT_TOKEN means CT, ~CURRENT_TOKEN means AT
    assign CURRENT_TOKEN = CANCEL_TOKEN_AVAILABLE;
  end
  endgenerate

  // ----- CURRENT_TOKEN_PSEUDO -----
  generate
  if (START_CTRL_IN ==0) begin
    // no token queue => no information about stored tokens
    assign CURRENT_TOKEN_PSEUDO = 1'b0;
  end

  if (START_CTRL_IN == 1) begin
    // pseudo token available (with cancel or activate token)
    assign CURRENT_TOKEN_PSEUDO = PSEUDO_TOKEN_AVAILABLE;
  end
  endgenerate


  // ----- CANCEL_STATE / CANCEL_STATE_CTRL HANDLING -----

  generate
  // a cancel might enter the module from its data successors only
  // cancels will be accepted until the cancel_counter is full
  // CANCEL STATE will be set according to the computation_counter
  if (START_CTRL_IN == 0 & STATIC_CT == 0) begin
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        CANCEL_ACK          <= 1'b1;
        cancel_state_keep      <= 1'b0;
        cancel_state_ctrl_keep <= 1'b0;
        cancel_to_process      <= 1'b0;
      end
      else begin
        if (cancel_counter != {(CANCEL_BITS){1'b1}}) begin
          CANCEL_ACK <= 1'b1;
        end
        else begin
          CANCEL_ACK <= 1'b0;
        end
      end
    end
  end

  // a cancel might enter the module from the controller
  // cancels are accepted until the ac token queue is full
  // cancel state ctrl will remain high until it has been processed
  // cancel state will remain high until it has been processed
  if (START_CTRL_IN == 1 & STATIC_CT == 0) begin
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        CANCEL_ACK          <= 1'b1;
        cancel_state_keep      <= 1'b0;
        cancel_state_ctrl_keep <= 1'b0;
        cancel_to_process      <= 1'b0;
      end
      else begin
        CANCEL_ACK     <= ~ac_token_queue_full;
        if (CANCEL_STATE_CTRL) begin
          // reset keep signal on handshake completion
          cancel_state_ctrl_keep <= ~CANCEL_STATE_CTRL_ACK;
        	// CANCEL_STATE_CTRL set but no read -> there is a unacknowledged CANCEL
        	// initially the keeper signal is zero, thus a read determines whether a cancel needs to be forwarded
          // as long as the keep signals remains set, cancel_to_process will not change its value
          // after both keep signals were reset the else tree will be evaluated
          cancel_to_process <= (cancel_to_process | ~cancel_state_ctrl_keep) & ~ac_token_queue_read;
        end
        else begin
          if (cancel_to_process & ac_token_queue_read) begin
             cancel_to_process  <= 1'b0;
          end
        end
      end
    end
  end

  // without control edge and static cancel tokens
  // a cancel will never enter this module
  if (START_CTRL_IN == 0 & STATIC_CT == 1) begin
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        CANCEL_ACK          <= 1'b0;
        cancel_state_keep      <= 1'b0;
        cancel_state_ctrl_keep <= 1'b0;
        cancel_to_process      <= 1'b0;
      end
    end
  end

  // a cancel might enter the module from the controller
  // cancels are accepted until the ac token queue is full
  // cancel state ctrl will remain high until it has been processed
  if (START_CTRL_IN == 1 & STATIC_CT == 1) begin
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        CANCEL_ACK          <= 1'b1;
        cancel_state_keep      <= 1'b0;
        cancel_state_ctrl_keep <= 1'b0;
        cancel_to_process      <= 1'b0;
      end
      else begin
        CANCEL_ACK     <= ~ac_token_queue_full;

        if (CANCEL_STATE) begin
          // reset keep signal on handshake completion
          cancel_state_keep      <= ~CANCEL_STATE_ACK;
        end

        if (CANCEL_STATE_CTRL) begin
          // reset keep signal on handshake completion
          cancel_state_ctrl_keep <= ~CANCEL_STATE_CTRL_ACK;
        end

        if (CANCEL_STATE | CANCEL_STATE_CTRL) begin

        	// CANCEL STATE or CANCEL_STATE_CTRL set but no read -> there is a unacknowledged CANCEL
        	// initially both keeper signals are zero, thus a read determines whether a cancel needs to be forwarded
          // as long as one of the keep signals remains set, cancel_to_process will not change its value
          // after both keep signals were reset the else tree will be evaluated
          cancel_to_process <= (cancel_to_process | ~(cancel_state_keep | cancel_state_ctrl_keep)) & ~ac_token_queue_read;
        end
        else begin
          // on ac_token_queue_read cancel has been processed
          if (cancel_to_process & ac_token_queue_read) begin
             cancel_to_process  <= 1'b0;
          end
        end
      end
    end
  end
  endgenerate


  // RESOURCES

  // ----- CANCEL COUNTER -----

  // the cancel counter takes care of the pending cancel requests
  // when a CANCEL enters the module and corresponding data can not be deleted
  // immediately or forwarding via CANCEL STATE needs to be initiated (dynamic cancel tokens)
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      cancel_counter <=  {(CANCEL_BITS){1'b0}};
    end else begin
      if (data_empty) begin
        case ({decrease_cancel_counter, increase_cancel_counter})
             2'b00: begin
                cancel_counter <= cancel_counter;
             end
             2'b01: begin
                if (cancel_counter != {(CANCEL_BITS){1'b1}}) begin
                   cancel_counter <= cancel_counter + 1'b1;
                end
             end
             2'b10: begin
                if (cancel_counter != {(CANCEL_BITS){1'b0}}) begin
                   cancel_counter <= cancel_counter - 1'b1;
                end
             end
             2'b11: begin
                  cancel_counter <= cancel_counter;
             end
        endcase
      end
    end
  end


  // ----- COMPUTATION COUNTER -----

  // the computation counter is required to determine whether a cancel needs to be forwared
  // via the cancel state output (dynamic cancel tokens only)
  // the coputation counter is also required in a special COMRADE case when there is a incomming control edge
  // since the data and memory dependencies both affect the START input CANCEL_STATE is used for the purpose to
  // allow the data to enter the module even if memory dependencies are unfulfilled
  generate
  if (START_CTRL_IN == 1 | STATIC_CT == 0) begin
    // count ongoing computations
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        computation_counter <=  {(COMPUTATION_BITS){1'b0}};
      end
      else begin
        case ({decrease_computation_counter, increase_computation_counter})
             2'b00: computation_counter <= computation_counter;
             2'b01: if (computation_counter != {(COMPUTATION_BITS){1'b1}})
                       computation_counter <= computation_counter + 1'b1;
             2'b10: if (computation_counter != {(COMPUTATION_BITS){1'b0}})
                       computation_counter <= computation_counter - 1'b1;
             2'b11: computation_counter <= computation_counter;
        endcase
      end
    end
  end
  else begin
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        computation_counter <=  {(COMPUTATION_BITS){1'b0}};
      end else begin
        computation_counter <= computation_counter;
      end
    end
  end
  endgenerate

  // ----- DATA_QUEUE INSTANTIATION -----
  generate
  if (DATA_SPECULATION == 0) begin
  disram_queue_non_spec #(.WR(WR), .QDEPTH(QDEPTH_INT), .QBITS(QBITS)) queue_out (
    .CLK(CLK),
    .RESET(RESET),
    .DIN(DATA_SHIFTREG_TO_QUEUE),
    .READ(data_read),
    .WRITE(data_write),
    .DOUT(DATA_QUEUE_TO_OUT),
    .EMPTY(data_empty),
    .FULL(data_full));
  end
  endgenerate

  // ----- TOKEN_QUEUE INSTANTIATION -----
  generate

  if(START_CTRL_IN == 1 & DATA_SPECULATION == 0) begin
  sr_queue #(.WR(1), .QDEPTH(TQDEPTH_INT), .QBITS(TQBITS)) ac_token_queue (
    .CLK(CLK),
    .RESET(RESET),
    .DIN(ac_token_queue_in),
    .READ(ac_token_queue_read),
    .WRITE(ac_token_queue_write),
    .DOUT(ac_token_queue_out),
    .EMPTY(ac_token_queue_empty),
    .FULL(ac_token_queue_full));
  end

  if(START_CTRL_IN == 1 & DATA_SPECULATION == 0 & PSEUDO_AT == 1) begin
  sr_queue #(.WR(1), .QDEPTH(TQDEPTH_INT), .QBITS(TQBITS)) ac_token_queue_pseudo (
    .CLK(CLK),
    .RESET(RESET),
    .DIN(pseudo_token_queue_in),
    .READ(ac_token_queue_read),
    .WRITE(ac_token_queue_write),
    .DOUT(pseudo_token_queue_out),
    .EMPTY(pseudo_token_queue_empty),
    .FULL(pseudo_token_queue_full));
  end

  endgenerate

  // make state of ac_token_queue visible during simulation
  // this is for debug purpose only
  `ifdef synthesis
  `else
  reg [15:0] ac_tokens_valid;
  reg [15:0] ac_tokens;
  integer empty_bit_position;
  generate

    if (START_CTRL_IN == 1) begin
      always @(posedge CLK, posedge RESET) begin
        if (RESET) begin
          ac_tokens_valid <= 0;
          ac_tokens <= 0;
          empty_bit_position <= 0;
        end
        else begin
          if (ac_token_queue_write & ~ac_token_queue_read) begin
            // write to queue, insert data at empty bit position, start to search at LSB
            empty_bit_position <= 0;
            while (ac_tokens_valid[empty_bit_position]) begin
              empty_bit_position = empty_bit_position + 1;
            end
            ac_tokens[empty_bit_position] <= ac_token_queue_in;
            ac_tokens_valid[empty_bit_position] <= 1'b1;
          end
          else if (~ac_token_queue_write & ac_token_queue_read) begin
            // read from queue, delete data at bit position 0 by right-shifting
            ac_tokens <= ac_tokens >> 1;
            ac_tokens_valid <= ac_tokens_valid >> 1;
          end
          else if (~ac_token_queue_write & ~ac_token_queue_read) begin
            // do nothing
          end
          else if (ac_token_queue_write & ac_token_queue_read) begin
            // if queue is empty, queue stays empty;
            // otherwise, first read datum, then write datum
            if (~ac_token_queue_empty) begin
              // read => shift
              ac_tokens <= ac_tokens >> 1;
              ac_tokens_valid <= ac_tokens_valid >> 1;
              // write
              empty_bit_position <= 0;
              while (ac_tokens_valid[empty_bit_position]) begin
                empty_bit_position = empty_bit_position + 1;
              end
              ac_tokens[empty_bit_position] <= ac_token_queue_in;
              ac_tokens_valid[empty_bit_position] <= 1'b1;
            end
          end
        end
      end
    end

    if (START_CTRL_IN == 0) begin
      // no ac_token_queue
      always @(posedge CLK, posedge RESET) begin
        ac_tokens_valid <= 0;
        ac_tokens <= 0;
        empty_bit_position <= 0;
      end
    end

  endgenerate
  `endif
