  // E.I.S. / ESA modlib for COMRADE and HardScale
  //
  // Description:
  // ----------------
  // Simplified multiplex operation (no CTs necessary) with output queue and shift register mechanism.
  // Just forwards active input; sequencer has to guarantee that there is only 1 input active at a time.
  //
  // Version History:
  // ----------------
  // 100826: BT: intial release

`timescale 1ns / 1ns
module irq_data_mux
         #(parameter
                      WA                   = 32,
                      NIN                  = 2,
                      WB                   = NIN,
                      WR                   = 32,
                      ONEHOT               = 1,
                      DEPTH                = 0,
                      QDEPTH               = 0,
                      TQDEPTH              = 0,
                      STATIC_CT            = 0,
                      START_CTRL_IN        = 0,
                      SIGN                 = 0,
                      PSEUDO_AT            = 0,
                      DATA_SPECULATION     = 0,
                      SPECULATION_BOUNDARY = 1,
                      NO_CT                = 0
          )(
           input  wire [(WA*NIN-1):0]    A,
           input  wire [(WB-1):0]        B,
           output wire [(WR-1):0]        R,
           input  wire                   RESET,
           input  wire                   CLK,
           input  wire                   CE,
           input  wire                   CONTROLLER_CANCEL,
           output wire                   TOKENS_EMPTY,
           output wire                   CURRENT_TOKEN,
           output wire                   CURRENT_TOKEN_PSEUDO,
           input  wire                   START,
           output wire                   START_ACK,
           input  wire                   START_CTRL,
           output wire                   START_CTRL_ACK,
           input  wire                   START_CTRL_PSEUDO,
           input  wire                   CANCEL,
           output reg                    CANCEL_ACK,
           output wire                   CANCEL_STATE,
           input  wire                   CANCEL_STATE_ACK,
           output wire                   CANCEL_STATE_CTRL,
           input  wire                   CANCEL_STATE_CTRL_ACK,
           output wire                   RESULT_READY,
           input  wire                   RESULT_READY_ACK);
           
  // define LATENCY of this module
  localparam LATENCY          = 1;
  // CANCEL BITS depending on LATENCY and shift register DEPTH
  localparam CANCEL_BITS      = min_1(ceil_log2(LATENCY+DEPTH-1));
  // QDEPTH_INT has to be at least 16
  localparam QDEPTH_INT       = min_16(QDEPTH);
  // set QBITS accordingly
  localparam QBITS            = (QDEPTH == 0) ? 0 : ceil_log2(QDEPTH_INT-1);
  // Sel/TQs
  localparam TQDEPTH_INT      = min_16(TQDEPTH);
  // set TQBITS accordingly
  localparam TQBITS           = (TQDEPTH == 0) ? 0 : ceil_log2(TQDEPTH_INT-1);
  // COMPUTATION_BITS depending  on LATENCY shift register DEPTH and queue QDEPTH
  localparam COMPUTATION_BITS = min_1(ceil_log2(LATENCY+DEPTH+QDEPTH_INT-1));
  localparam NIN_BITS         = min_1(ceil_log2(NIN-1));

  integer i, j;
  
  `include "handshaking_assigns_verilog_only.v"
  
  reg [WR-1:0] r_int; // C/L

  // no control edge
  initial begin
    if (START_CTRL_IN == 1) begin
      $display("ERROR: START_CTRL_IN == 1 for irq_data_mux; must be 0.");
      $finish;
    end
  end

  // the operation
  generate
    always @(A, B) begin
      // initialize to avoid latches
      for (j=0; j < WR; j = j + 1) begin
        r_int[j] = A[j];
      end

      if (ONEHOT == 0) begin
        for (i=0; i < WR; i = i + 1) begin
          r_int[i] = A[B*WA+i];
        end
      end
      else begin
        for (i=0; i < NIN; i = i + 1) begin
          if (B[i]) begin
            for (j=0; j < WR; j = j + 1) begin
              r_int[j] = A[i*WA+j];
            end
          end
        end
      end
    end
  endgenerate


  // shift register instantiation
  shiftreg #(.WA(WR), .WR(WR), .DEPTH(DEPTH)) sr (
    .A(r_int),
    .R(DATA_SHIFTREG_TO_QUEUE),
    .RESET(RESET),
    .CLK(CLK),
    .CE(CE_SHIFTREG),
    .START(START_SHIFTREG),
    .RESULT_READY(RESULT_READY_SHIFTREG));


  `include "functions.v"

endmodule
