// E.I.S. / ESA modlib for COMRADE and HardScale
//
// Description:
// ----------------
// Multiplexer for loops: has dedicated loop initialization and continue inputs (init
// input is A[(WA-1):0], continue input is A[(WA*2-1):WA];
// with output queue and shift register mechanism
//
// Version History:
// ----------------
// 100826: BT: intial release

`timescale 1ns / 1ns

module loop_mux

          #(parameter WA                   = 32, 
                      NIN                  = 2,
                      WB                   = NIN, 
                      WR                   = 32,
                      ONEHOT               = 1,
                      DEPTH                = 0,
                      QDEPTH               = 16,
                      TQDEPTH              = 16,
                      STATIC_CT            = 0,
                      START_CTRL_IN        = 1,
                      SIGN                 = 0,
                      PSEUDO_AT            = 0,
                      DATA_SPECULATION     = 0,
                      SPECULATION_BOUNDARY = 0,
                      NO_CT                = 0)

          (input  wire [(WA*NIN-1):0]    A, 
           input  wire [(WB-1):0]        B, 
           output wire [(WR-1):0]        R,
           input  wire                   RESET,
           input  wire                   CLK,
           input  wire                   CE,
           input  wire                   CONTROLLER_CANCEL,
           output wire                   TOKENS_EMPTY,
           output wire                   CURRENT_TOKEN,
           output wire                   CURRENT_TOKEN_PSEUDO,
           input  wire [NIN-1:0]         START,
           output wire [NIN-1:0]         START_ACK,
           input  wire                   START_CTRL,
           output wire                   START_CTRL_ACK,
           input  wire                   CANCEL,
           output wire                   CANCEL_ACK,
           output reg  [NIN-1:0]         CANCEL_STATE,
           input  wire [NIN-1:0]         CANCEL_STATE_ACK,
           output wire                   CANCEL_STATE_CTRL,
           input  wire                   CANCEL_STATE_CTRL_ACK,
           output wire                   RESULT_READY,
           input  wire                   RESULT_READY_ACK);


  // define LATENCY of this module         
  localparam LATENCY          = 1;
  // Sel/TQs
  localparam TQDEPTH_INT      = min_16(TQDEPTH);
  localparam TQBITS           = (TQDEPTH == 0) ? 0 : ceil_log2(TQDEPTH_INT-1);
  // QDEPTH_INT has to be at least 16
  localparam QDEPTH_INT       = min_16(QDEPTH);
  // set QBITS accordingly
  localparam QBITS            = (QDEPTH == 0) ? 0 : ceil_log2(QDEPTH_INT-1);
  // COMPUTATION_BITS depending  on LATENCY shift register DEPTH and queue QDEPTH
  localparam COMPUTATION_BITS = min_1(ceil_log2(LATENCY+DEPTH+QDEPTH_INT-1));
  localparam NIN_BITS         = min_1(ceil_log2(NIN-1));

  `include "handshaking_assigns_loop_mux.v"
  
  // the operation
  assign r_int = START[0] & START_ACK[0] ? A[WA-1:0]    // init
               : START[1] & START_ACK[1] ? A[2*WA-1:WA] // continue
               : 0; // dummy
  
  // shift register instantiation    
  shiftreg #(.WA(WR), .WR(WR), .DEPTH(DEPTH)) sr (
    .A(r_int), 
    .R(DATA_SHIFTREG_TO_QUEUE), 
    .RESET(RESET), 
    .CLK(CLK), 
    .CE(CE_SHIFTREG),
    .START(START_SHIFTREG),
    .RESULT_READY(RESULT_READY_SHIFTREG)
  );
    
  `include "functions.v"

endmodule
