// E.I.S. / ESA modlib for COMRADE and HardScale
  //
  // TODO: Parameter QDEPTH currently controls the size of the (undocumented) address input
  //       queue "queue_addr" _and_ the (documented) data output queue "queue_data".
  //       To be able to use a data output queue _without_ an address
  //       input queue, there should be an additional parameter IN_QDEPTH which controls
  //       the address input queue size separately.
  //
  // Version History:
  // ----------------
  //
  // 100826: BT: intial release

`timescale 1ns / 1ns

module memread

          #(parameter PORTNUM        = 0,
                      WIDTH          = 2'b00, // 00 = 32 bit, 10 = 16 bit, 01 = 8 bit
                      DEPTH          = 1, // buffer read values before outputting them, to minimize critical path
                      QDEPTH         = 1,
                      TQDEPTH        = 1,
                      ADDR_WIDTH     = 24,
                      DATA_WIDTH     = 32,
                      RESULT_WIDTH   = (WIDTH == 0) ? 32 : (WIDTH * 8),
                      STATIC_CT      = 0,
                      START_CTRL_IN  = 1,
                      DATA_SPECULATION = 0,
                      SPECULATION_BOUNDARY = 0,
                      NUM_CACHEPORTS = 1, // WARNING: PSEUDO_AT functionality not yet implemented for memread! => always set PSEUDO_AT=0
                      PSEUDO_AT      = 0)

          (input  wire [ADDR_WIDTH-1:0]                  A,
           output wire [RESULT_WIDTH-1:0]                R,
           input  wire                                   RESET,
           input  wire                                   CLK,
           input  wire                                   CE,
           input  wire                                   CONTROLLER_CANCEL,
           output wire                                   TOKENS_EMPTY, // COCOMA needs to know if there are still tokens which have to be processed
           output wire                                   CURRENT_TOKEN,
           output wire                                   CURRENT_TOKEN_PSEUDO,
           input  wire                                   START,
           output wire                                   START_ACK,
           input  wire                                   START_CTRL,
           output wire                                   START_CTRL_ACK,
           input  wire                                   START_CTRL_PSEUDO,
           input  wire                                   CANCEL,
           output reg                                    CANCEL_ACK,
           output wire                                   CANCEL_STATE,
           input  wire                                   CANCEL_STATE_ACK,
           output wire                                   CANCEL_STATE_CTRL,
           input  wire                                   CANCEL_STATE_CTRL_ACK,

           output wire                                   RESULT_READY,
           input  wire                                   RESULT_READY_ACK,
           output wire                                   MEM_READY,
           input  wire                                   MEM_READY_ACK,

           output wire [(NUM_CACHEPORTS*ADDR_WIDTH)-1:0] CACHE_ADDR,
           output wire [(NUM_CACHEPORTS*2)-1:0]          CACHE_WIDTH,
           input  wire [(NUM_CACHEPORTS*DATA_WIDTH)-1:0] CACHE_READ,
           output wire [NUM_CACHEPORTS-1:0]              CACHE_OE,
           input  wire [NUM_CACHEPORTS-1:0]              CACHE_STALL
           );


  // QDEPTH_INT has to be at least 16
  localparam QDEPTH_INT       = min_16(QDEPTH);
  // CANCEL BITS depending on LATENCY and shift register DEPTH
  localparam CANCEL_BITS      = min_1(ceil_log2(QDEPTH_INT-1));
  // set QBITS accordingly
  localparam QBITS            = (QDEPTH == 0) ? 0 : ceil_log2(QDEPTH_INT-1);
  // TQDEPTH_INT has to be at least 16
  localparam TQDEPTH_INT      = min_16(TQDEPTH);
  // set TQBITS accordingly
  localparam TQBITS           = (TQDEPTH == 0) ? 0 : ceil_log2(TQDEPTH_INT-1);
  // COMPUTATION_BITS depending  on LATENCY shift register DEPTH and queue QDEPTH
  localparam COMPUTATION_BITS = (QDEPTH == 0) ? 1 : min_1(ceil_log2(QDEPTH_INT-1));

  wire pseudo_data_commit;
  wire ce_sr;

  // wire definition pseudo token queue
  wire pseudo_token_queue_in;
  wire pseudo_token_queue_out;
  
  // wire definition ac token queue
  wire ac_token_queue_in;
  wire ac_token_queue_out;
  wire ac_token_queue_out_pseudo;
  wire addr_write;
  wire addr_read;
  wire data_write;
  wire data_read;
  wire addr_full;
  wire addr_empty;
  wire data_full;
  wire data_empty;
  wire OPERATION;
  wire write_to_shiftreg;
  wire RESULT_READY_SHIFTREG;
  wire ac_token_queue_empty;
  wire pseudo_token_queue_empty;
  wire ac_token_queue_full;
  wire pseudo_token_queue_full;
  wire ac_token_queue_read;
  wire ac_token_queue_write;
  

  // mem_ready_count
  reg  [(COMPUTATION_BITS - 1):0] mem_ready_count;
  wire increase_mem_ready_count;
  wire decrease_mem_ready_count;

  wire [(ADDR_WIDTH - 1):0] DIN_Q_OUT;
  wire [(ADDR_WIDTH - 1):0] DOUT_Q_OUT;
  wire [(RESULT_WIDTH - 1):0] DATA_SHIFTREG_TO_QUEUE;
  wire [(RESULT_WIDTH - 1):0] DATA_QUEUE_TO_OUT;
  wire [(DATA_WIDTH - 1):0] DATA_CACHE_TO_SHIFTREG;
  wire [(ADDR_WIDTH - 1):0] ADDR_Q;

  reg  access_done;
  reg  CACHE_STALL_REG;
  reg  CACHE_OE_REG;
  reg  REQUEST_RUNNING;

  reg  cancel_state_keep;
  reg  cancel_state_ctrl_keep;
  reg  cancel_state_incomplete;
  reg  free_register;
  reg  cancel_register;

`ifdef synthesis
`else
  reg  [COMPUTATION_BITS:0] computation_counter;
  reg  [CANCEL_BITS:0]      cancel_counter;
`endif

  // an AT is consumed from the input interface
  wire ACTIVATE_CONSUMED = START_CTRL & START_CTRL_ACK;
  // a cancel request is accepted
  wire CANCEL_CONSUMED = CANCEL & CANCEL_ACK;
  // input data is consumed from the input interface
  wire INPUT_DATA_CONSUMED = START & START_ACK;
  // a result is ready and consumed
  wire RESULT_CONSUMED = RESULT_READY & RESULT_READY_ACK;
  // ac_token_queue currently has a CT at its output
  wire ac_token_queue_ct = ~ac_token_queue_empty & ac_token_queue_out;
  // ac_token_queue currently has an AT at its output
  wire ac_token_queue_at = ~ac_token_queue_empty & ~ac_token_queue_out;
  // StaticCT: delete CT from ac_token_queue,
  //           at the same time read an address from the address queue,
  //           and put a dummy value into the shiftreg
  wire create_pseudo_at_from_ct_and_addr_read;
  generate
    if (QDEPTH_INT != 0) begin
      assign create_pseudo_at_from_ct_and_addr_read = addr_read & ac_token_queue_ct & ~CACHE_OE_REG & ~access_done;
    end
    if (QDEPTH_INT == 0) begin
      assign create_pseudo_at_from_ct_and_addr_read = INPUT_DATA_CONSUMED & ac_token_queue_ct & ~CACHE_OE_REG & ~access_done;
    end
  endgenerate
  // DynamicCT: delete CT from ac_token_queue,
  //            at the same time read an address from the address queue or propagate the CT to the predecessor modules,
  //            and put a dummy value into the shiftreg
  wire create_pseudo_at_from_ct;
  generate
    if (QDEPTH_INT != 0) begin
      assign create_pseudo_at_from_ct = (addr_read | (CANCEL_STATE & CANCEL_STATE_ACK)) & ac_token_queue_ct & ~CACHE_OE_REG & ~access_done;
    end
    if (QDEPTH_INT == 0) begin
      assign create_pseudo_at_from_ct = (INPUT_DATA_CONSUMED | (CANCEL_STATE & CANCEL_STATE_ACK)) & ac_token_queue_ct & ~CACHE_OE_REG & ~access_done;
    end
  endgenerate

  // ac_token_queue_write validates ac_token_queue_in on token arrival
  assign ac_token_queue_write  = CANCEL_CONSUMED || ACTIVATE_CONSUMED;

  generate
  if (PSEUDO_AT == 0) begin
    assign pseudo_token_queue_out = 1'b0;
  end
  endgenerate
     

  // ctrl_cmd == 1 means CANCEL, ctrl_cmd == 0 means ACTIVATE
  assign ac_token_queue_in = CANCEL_CONSUMED;

  // there is a valid result at the output, if the output queue is filled
  // or a computation just completed
  assign RESULT_READY = (~data_empty | RESULT_READY_SHIFTREG);
  assign ce_sr = ((QDEPTH_INT > 0) & ~data_full) | data_read |
                 (~RESULT_READY & ((START_CTRL_IN == 0) | ~RESULT_READY_SHIFTREG | ~ac_token_queue_empty | ACTIVATE_CONSUMED));


  generate
  if (DATA_SPECULATION == 0) begin
    assign data_read_conditional    = data_read;
  end
  endgenerate
  
  
  // SINGLE ACCESS TOKEN QUEUE
  // cancel_register / free register
  generate
  if (STATIC_CT == 1 & QDEPTH_INT == 0) begin
    // keep operation on stall
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        cancel_register <= 1'b0;
      end
      else begin
        if (CANCEL_CONSUMED) begin
          cancel_register <= 1'b1;
        end
        else if (INPUT_DATA_CONSUMED) begin
          cancel_register <= 1'b0;
         end
      end
    end
  end

  if (STATIC_CT == 0 & QDEPTH_INT == 0) begin
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        cancel_register <= 1'b0;
      end
      else begin
         if (CANCEL_CONSUMED) begin
           cancel_register <= 1'b1;
         end
         else if (INPUT_DATA_CONSUMED | (CANCEL_STATE & CANCEL_STATE_ACK)) begin
           cancel_register <= 1'b0;
         end
      end
    end
  end

  if (QDEPTH_INT == 0) begin
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        free_register   <= 1'b1;
      end
      else begin
        if (INPUT_DATA_CONSUMED & ACTIVATE_CONSUMED) begin
          free_register <= 1'b0;
        end
        else if (RESULT_CONSUMED) begin
          free_register <= 1'b1;
        end
      end
    end
  end
  endgenerate


  generate
  if (STATIC_CT == 1 & QDEPTH_INT > 0) begin
    // a read request is accepted, when the buffers are not full
    assign START_ACK = START & ~addr_full & ~data_full;  
  end                          
  if (STATIC_CT == 0 & QDEPTH_INT > 0) begin
    // a read request is accepted, when the buffers are not full
    // do not accept data when CANCEL STATE is sent
    assign START_ACK = START & ~addr_full & ~data_full & ~CANCEL_STATE;
  end
  if (QDEPTH_INT == 0) begin
    assign START_ACK = START & ((START_CTRL & ~CACHE_STALL[PORTNUM] & free_register) | cancel_register);
  end
  endgenerate
  
  generate
  if (QDEPTH_INT == 0) begin
    assign START_CTRL_ACK = INPUT_DATA_CONSUMED & START_CTRL;
  end
  if (QDEPTH_INT > 0) begin
    // set when there is space left in the activate/cancel token queue
    assign START_CTRL_ACK = START_CTRL & (~ac_token_queue_full);
  end
  endgenerate
    
   
  generate
  if (QDEPTH_INT == 0) begin
    // when the operands and the AT are available, the operation can be started
    assign OPERATION = INPUT_DATA_CONSUMED & ((ac_token_queue_empty & ACTIVATE_CONSUMED) | ac_token_queue_at);
  end
  if (QDEPTH_INT > 0 & DATA_SPECULATION == 0) begin
    // when the operands and the AT are available, the operation can be started
    assign OPERATION = (~addr_empty | INPUT_DATA_CONSUMED) & ((ac_token_queue_empty & ACTIVATE_CONSUMED) | ac_token_queue_at);
  end
  endgenerate


  generate
  if (STATIC_CT == 0) begin
    // CANCEL_STATE_CTRL is always zero
    assign CANCEL_STATE_CTRL  = 1'b0;
    // delete token when corresponding element is removed from queue
    // or whenever data is transferred or cancelled
    // ac_token_queue_read is executed whenever ac_token_queue_write is active and the data should not be written
    assign ac_token_queue_read    = addr_read | (addr_empty & INPUT_DATA_CONSUMED & ~CACHE_STALL[PORTNUM] & ac_token_queue_at)
                               | (addr_empty & INPUT_DATA_CONSUMED & ~CACHE_STALL[PORTNUM] &  ac_token_queue_empty & ACTIVATE_CONSUMED)
                               | create_pseudo_at_from_ct;
  end   
  if (STATIC_CT == 1) begin
    // CANCEL_STATE_CTRL = 1, if there is a CT stored in this node
    assign CANCEL_STATE_CTRL  = (ac_token_queue_ct & ~cancel_state_incomplete) | cancel_state_ctrl_keep;
    // delete token when corresponding element removed from queue
    // or data just gets in and is processed with token
    assign ac_token_queue_read    = (addr_read & ac_token_queue_at) // read because AT is consumed
                               | (addr_read & ac_token_queue_empty) // read token if simultaneous write token in token consumption case
                               | (addr_empty & INPUT_DATA_CONSUMED & ~CACHE_STALL[PORTNUM] & ac_token_queue_at)
                               | (addr_empty & INPUT_DATA_CONSUMED & ~CACHE_STALL[PORTNUM] &  ac_token_queue_empty & ACTIVATE_CONSUMED)
                               | create_pseudo_at_from_ct_and_addr_read;
  end
  endgenerate


  generate  
  // a read request is accepted, when the buffers are not full
  // there is no CANCEL by definition of the controller
  if (STATIC_CT == 1 & QDEPTH_INT > 0) begin
    // CANCEL_STATE = 1, if there is a CT stored in this node
    assign CANCEL_STATE  = (ac_token_queue_ct & ~cancel_state_incomplete) | cancel_state_keep;
  end
  if (STATIC_CT == 1 & QDEPTH_INT == 0) begin
    assign CANCEL_STATE       = cancel_register & free_register;
  end                            
  if (STATIC_CT == 0 & QDEPTH_INT > 0) begin
    // there is a cancel in the ac_queue but no corresponding element to cancel
    assign CANCEL_STATE = ac_token_queue_ct & addr_empty;
  end
  if (STATIC_CT == 0 & QDEPTH_INT == 0) begin
    assign CANCEL_STATE       = cancel_register & free_register;
  end
  endgenerate


  // assert CANCEL_STATE on CANCEL request
  // in case no computation is ongoin
  // deassert CANCEL_STATE on CANCEL_STATE_ACK input
  generate
  if (START_CTRL_IN == 0 & STATIC_CT == 0) begin
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      CANCEL_ACK     <= 1'b1;
    end
    else begin
      $display("ERROR: START_CTRL_IN has to be 1 for memread");
      $finish;
    end
  end
  end

  // no cancel and thus no cancel accept
  if (START_CTRL_IN == 0 & STATIC_CT == 1) begin
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      CANCEL_ACK     <= 1'b0;
    end
    else begin
      $display("ERROR: START_CTRL_IN has to be 1 for memread");
      $finish;
    end
  end
  end

  if (START_CTRL_IN == 1 & STATIC_CT == 0) begin
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
        CANCEL_ACK           <= 1'b1;
        cancel_state_ctrl_keep  <= 1'b0;
        cancel_state_incomplete <= 1'b0;
    end
    else begin
      if (QDEPTH_INT == 0) begin
        CANCEL_ACK     <= ~cancel_register;
      end
      if (QDEPTH_INT > 0) begin
        CANCEL_ACK     <= ~ac_token_queue_full;
      end
      if (CANCEL_STATE_CTRL) begin
        // reset keep signal on handshake completion
        cancel_state_ctrl_keep <= ~CANCEL_STATE_CTRL_ACK;
        // initially keep signals are zero, thus a read determines whether there is something to process
        // as long as the keep signal is set an initial zero will remain zero, however an initial one can be set to zero
        // when keep goes low the condition is not fullfilled anymore
        cancel_state_incomplete <= (cancel_state_incomplete | ~(cancel_state_ctrl_keep)) & ~ac_token_queue_read;
        CANCEL_ACK     <= 1'b0;
      end
      else if (cancel_state_incomplete & ac_token_queue_read) begin
          cancel_state_incomplete  <= 1'b0;
      end
    end
  end
  end


  if (START_CTRL_IN == 1 & STATIC_CT == 1) begin
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
        CANCEL_ACK           <= 1'b1;
        cancel_state_keep       <= 1'b0;
        cancel_state_ctrl_keep  <= 1'b0;
        cancel_state_incomplete <= 1'b0;
    end
    else begin
      if (QDEPTH_INT == 0) begin
        CANCEL_ACK     <= ~cancel_register;
      end

      if (QDEPTH_INT > 0) begin
        CANCEL_ACK     <= ~ac_token_queue_full;
      end

      if (CANCEL_STATE) begin
        // reset keep signal on handshake completion
        cancel_state_keep      <= ~CANCEL_STATE_ACK;
      end

      if (CANCEL_STATE_CTRL) begin
        // reset keep signal on handshake completion
        cancel_state_ctrl_keep <= ~CANCEL_STATE_CTRL_ACK;
      end

      if (CANCEL_STATE | CANCEL_STATE_CTRL) begin
        // initially keep signals are zero, thus a read determines whether there is something to process
        // as long as the keep signal is set an initial zero will remain zero, however an initial one can be set to zero
        // when keep goes low the condition is not fullfilled anymore
        cancel_state_incomplete <= (cancel_state_incomplete | ~(cancel_state_keep | cancel_state_ctrl_keep)) & ~ac_token_queue_read;
      end
      // on ac_token_queue_read cancel has been processed
      else if (cancel_state_incomplete & ac_token_queue_read) begin
          cancel_state_incomplete  <= 1'b0;
      end
    end
  end
  end
  endgenerate



`ifdef synthesis
`else
  // visualization counters

  wire decrease_cancel_counter;
  wire increase_cancel_counter;
  wire decrease_computation_counter;
  wire increase_computation_counter;

  // count ongoing computations
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      computation_counter <=  {(COMPUTATION_BITS+1){1'b0}};
    end else begin
      case ({decrease_computation_counter, increase_computation_counter})
           2'b00: computation_counter <= computation_counter;
           2'b01: if (computation_counter != {1'b1, {(COMPUTATION_BITS){1'b0}}}) // max value for computation_counter
                 computation_counter <= computation_counter + 1'b1;
           2'b10: if (computation_counter != {(COMPUTATION_BITS+1){1'b0}})
                 computation_counter <= computation_counter - 1'b1;
           2'b11: computation_counter <= computation_counter;
      endcase
    end
  end

  generate
  if (STATIC_CT == 0) begin
    assign increase_cancel_counter      = CANCEL_CONSUMED;
    assign increase_computation_counter = INPUT_DATA_CONSUMED;
  end
  if (STATIC_CT == 1) begin
    assign increase_cancel_counter      = CANCEL_CONSUMED;
    assign increase_computation_counter = INPUT_DATA_CONSUMED && ~(~ac_token_queue_empty & addr_empty & ac_token_queue_out);
  end
  endgenerate
  
  generate    
  if (STATIC_CT == 1 & QDEPTH_INT > 0) begin
    assign decrease_cancel_counter      = ac_token_queue_read & ac_token_queue_ct;
    assign decrease_computation_counter = data_read;
  end
  if (STATIC_CT == 1 & QDEPTH_INT == 0) begin
    assign decrease_cancel_counter      = CANCEL_CONSUMED;
    assign decrease_computation_counter = data_read | CANCEL_CONSUMED;
  end
  
  if (STATIC_CT == 0 & QDEPTH_INT > 0) begin
    assign decrease_cancel_counter      = ac_token_queue_read & ac_token_queue_ct;
    assign decrease_computation_counter = data_read;
  end
  if (STATIC_CT == 0 & QDEPTH_INT == 0) begin
    assign decrease_cancel_counter      = CANCEL_CONSUMED;
    assign decrease_computation_counter = data_read | CANCEL_CONSUMED;
  end
  endgenerate


  // count non-extinguished CTs
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      cancel_counter <=  {(CANCEL_BITS+1){1'b0}};
    end else begin
        // decrease: delete address from address queue or from input
        // increase: CT comes in
        case ({decrease_cancel_counter, increase_cancel_counter})
             2'b00: cancel_counter <= cancel_counter;
             2'b01: if (cancel_counter != {1'b1, {(CANCEL_BITS){1'b0}}}) // max value for cancel_counter
                   cancel_counter <= cancel_counter + 1'b1;
             2'b10: if (cancel_counter != {(CANCEL_BITS+1){1'b0}})
                   cancel_counter <= cancel_counter - 1'b1;
             2'b11: cancel_counter <= cancel_counter;
        endcase
    end
  end
`endif

  // MARC has completed a read access, if the current value from the cache is a valid read datum;
  // this is the case if a value was requested from the cache during the previous cycle,
  // there was no stall in the previous cycle, and there was a request in the user-buffer in the previous cycle already
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      //ac_valid <= 1'b0;
      access_done <= 1'b0;
    end
    else begin
      //ac_valid <= ac_token_queue_read;
      access_done <= ~CACHE_STALL[PORTNUM] & REQUEST_RUNNING;
    end
  end

generate
  // dummy read on cancel
  if (STATIC_CT == 1) begin
    assign write_to_shiftreg = access_done |
                               create_pseudo_at_from_ct_and_addr_read;
  end
  if (STATIC_CT == 0) begin
    assign write_to_shiftreg = access_done |
                               create_pseudo_at_from_ct;
  end
endgenerate

  // write address/data to queue, if a new request is not directly consumed by MARC
  assign addr_write = INPUT_DATA_CONSUMED & ~(ac_token_queue_at & addr_empty & ~CACHE_STALL[PORTNUM])
                                          & ~(ac_token_queue_empty & addr_empty & ACTIVATE_CONSUMED & ~CACHE_STALL[PORTNUM]);


  // read (= remove) current element from address queue when either the current address is now successfully sent to MARC
  // or a cancellation request is pending
  assign addr_read = ~addr_empty & ((CACHE_OE[PORTNUM] & ~CACHE_STALL[PORTNUM]) | ac_token_queue_ct);

  // write data into the data output queue, if an access is finished, but the data cannot be consumed
  // by all successor nodes currently
  assign data_write = RESULT_READY_SHIFTREG & (~RESULT_CONSUMED | ~data_empty);

  // read data from the data output queue, if the successor consumes a result (cancellation is not allowed here)
  assign data_read  = RESULT_CONSUMED;


  assign DIN_Q_OUT              = A;
  assign ADDR_Q                 = DOUT_Q_OUT;
  assign DATA_CACHE_TO_SHIFTREG = CACHE_READ[((PORTNUM+1)*DATA_WIDTH-1):(PORTNUM*DATA_WIDTH)];

  assign R         = data_empty ? DATA_SHIFTREG_TO_QUEUE : DATA_QUEUE_TO_OUT;


  // output assignment - set unused ports to 0 to prevent 'z' or 'x' values on the cache buses
generate
  // NUM_CACHEPORTS-1 .. PORTNUM+1
  if (PORTNUM < NUM_CACHEPORTS - 1) begin
    assign CACHE_ADDR   [NUM_CACHEPORTS*ADDR_WIDTH-1:(PORTNUM+1)*ADDR_WIDTH] = 0;
    assign CACHE_WIDTH  [NUM_CACHEPORTS*2         -1:(PORTNUM+1)*2         ] = 0;
    assign CACHE_OE     [NUM_CACHEPORTS           -1:(PORTNUM+1)           ] = 0;
  end
  // PORTNUM (actual assignments)
  assign CACHE_ADDR [ (PORTNUM+1)*ADDR_WIDTH-1 : PORTNUM*ADDR_WIDTH ] = OPERATION ? (addr_empty ? A : ADDR_Q) : 0;
  assign CACHE_WIDTH[((PORTNUM+1)*2         -1):(PORTNUM*2         )] = OPERATION ? WIDTH : 0;
  assign CACHE_OE   [  PORTNUM                                      ] = OPERATION;

  // PORTNUM-1 .. 0
  if (PORTNUM > 0) begin
    assign CACHE_ADDR [PORTNUM*ADDR_WIDTH -1:0] = 0;
    assign CACHE_WIDTH[PORTNUM*2          -1:0] = 0;
    assign CACHE_OE   [PORTNUM            -1:0] = 0;
  end
endgenerate

  // REQUEST_RUNNING
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      REQUEST_RUNNING <= 1'b0;
    end else begin
      // start new request, or keep existing request
      REQUEST_RUNNING <= (CACHE_OE[PORTNUM] && ~CACHE_STALL[PORTNUM]) || (REQUEST_RUNNING && CACHE_STALL[PORTNUM]);
    end
  end

  // CACHE_STALL_REG
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      CACHE_STALL_REG <= 1'b0;
    end else begin
      CACHE_STALL_REG <= CACHE_STALL[PORTNUM];
    end
  end

  // CACHE_OE_REG
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      CACHE_OE_REG <= 1'b0;
    end else begin
      CACHE_OE_REG <= CACHE_OE[PORTNUM];
    end
  end


  // mem_ready_count: buffers AT going out via MEM_READY
  assign increase_mem_ready_count = OPERATION & ~CACHE_STALL[PORTNUM];
  assign decrease_mem_ready_count = MEM_READY_ACK;
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      mem_ready_count <= 0;
    end else begin
      case ({decrease_mem_ready_count, increase_mem_ready_count})
           2'b00: mem_ready_count <= mem_ready_count;
           2'b01: if (mem_ready_count != {(COMPUTATION_BITS){1'b1}}) // max value for mem_ready_count
                 mem_ready_count <= mem_ready_count + 1'b1;
           2'b10: if (mem_ready_count != {(COMPUTATION_BITS){1'b0}})
                 mem_ready_count <= mem_ready_count - 1'b1;
           2'b11: mem_ready_count <= mem_ready_count;
      endcase
    end
  end

  assign MEM_READY = (mem_ready_count != {(COMPUTATION_BITS){1'b0}});

  generate
  if (DATA_SPECULATION == 0) begin
  // (1) queue for addresses
  disram_queue_non_spec #(.WR(ADDR_WIDTH), .QDEPTH(QDEPTH_INT), .QBITS(QBITS)) queue_addr (
    .CLK(CLK),
    .RESET(RESET),
    .DIN(DIN_Q_OUT),
    .READ(addr_read),
    .WRITE(addr_write),
    .DOUT(DOUT_Q_OUT),
    .EMPTY(addr_empty),
    .FULL(addr_full));
  end
  endgenerate
  // (2) after that, request is sent to MARC

  // (3) read values are stored in a shift register to minimize the critical path
  shiftreg #(.WA(RESULT_WIDTH), .WR(RESULT_WIDTH), .DEPTH(DEPTH)) sr (
    .A(DATA_CACHE_TO_SHIFTREG[RESULT_WIDTH-1:0]),
    .R(DATA_SHIFTREG_TO_QUEUE),
    .RESET(RESET),
    .CLK(CLK),
    .CE(ce_sr),
    .START(write_to_shiftreg),
    .RESULT_READY(RESULT_READY_SHIFTREG));


  generate
  if (DATA_SPECULATION == 0) begin
  disram_queue_non_spec #(.WR(RESULT_WIDTH), .QDEPTH(QDEPTH_INT), .QBITS(QBITS)) queue_out (
    .CLK(CLK),
    .RESET(RESET),
    .DIN(DATA_SHIFTREG_TO_QUEUE),
    .READ(data_read),
    .WRITE(data_write),
    .DOUT(DATA_QUEUE_TO_OUT),
    .EMPTY(data_empty),
    .FULL(data_full));
  end
  endgenerate

  generate
  if(START_CTRL_IN == 1 & DATA_SPECULATION == 0 & PSEUDO_AT == 1) begin
  // pseudoAT queue stores a "1" per token of ac_token_queue if that token is a pseudo AT, a "0" otherwise
  sr_queue #(.WR(1), .QDEPTH(TQDEPTH_INT), .QBITS(TQBITS)) ac_token_queue_pseudo (
    .CLK(CLK),
    .RESET(RESET),
    .DIN(pseudo_token_queue_in),
    .READ(ac_token_queue_read),
    .WRITE(ac_token_queue_write),
    .DOUT(pseudo_token_queue_out),
    .EMPTY(pseudo_token_queue_empty),
    .FULL(pseudo_token_queue_full));
  end
    
  if(START_CTRL_IN == 1 & DATA_SPECULATION == 0) begin
  // queue to keep activate and cancel tokens
  sr_queue #(.WR(1), .QDEPTH(TQDEPTH_INT), .QBITS(TQBITS)) ac_token_queue (
    .CLK(CLK),
    .RESET(RESET),
    .DIN(ac_token_queue_in),
    .READ(ac_token_queue_read),
    .WRITE(ac_token_queue_write),
    .DOUT(ac_token_queue_out),
    .EMPTY(ac_token_queue_empty),
    .FULL(ac_token_queue_full));
  end
  endgenerate

  // ----- TOKENS_EMPTY -----
  generate
  if (START_CTRL_IN == 1) begin
    assign TOKENS_EMPTY = ac_token_queue_empty;
  end else begin
    assign TOKENS_EMPTY = 1'b1;
  end
  endgenerate

  // ----- CURRENT_TOKEN -----
  generate
  if (START_CTRL_IN == 1) begin
    assign CURRENT_TOKEN = ac_token_queue_ct; // CURRENT_TOKEN means CT, ~CURRENT_TOKEN means AT
  end else begin
    assign CURRENT_TOKEN = 1'b0; // no token queue => no information about stored tokens
  end
  endgenerate

  // ----- CURRENT_TOKEN_PSEUDO -----
  generate
  if (START_CTRL_IN == 1) begin
    assign CURRENT_TOKEN_PSEUDO = ~ac_token_queue_empty & ac_token_queue_out_pseudo;
  end else begin
    assign CURRENT_TOKEN_PSEUDO = 1'b0; // no token queue => no information about stored tokens
  end
  endgenerate

  `include "functions.v"


// make state of ac_token_queue visible during simulation
`ifdef synthesis
`else
   reg [15:0] ac_tokens_valid;
   reg [15:0] ac_tokens;
   integer    empty_bit_position;

    generate
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        ac_tokens_valid <= 0;
        ac_tokens <= 0;
        empty_bit_position = 0;
      end else begin
        if (ac_token_queue_write & ~ac_token_queue_read) begin
          // write to queue, insert data at empty bit position, start to search at LSB
          empty_bit_position = 0;
          while (ac_tokens_valid[empty_bit_position]) begin
            empty_bit_position = empty_bit_position + 1;
          end
          ac_tokens[empty_bit_position] <= ac_token_queue_in;
          ac_tokens_valid[empty_bit_position] <= 1'b1;
        end else if (~ac_token_queue_write & ac_token_queue_read) begin
          // read from queue, delete data at bit position 0 by right-shifting
          ac_tokens <= ac_tokens >> 1;
          ac_tokens_valid <= ac_tokens_valid >> 1;
        end else if (~ac_token_queue_write & ~ac_token_queue_read) begin
          // do nothing
        end else if (ac_token_queue_write & ac_token_queue_read) begin
          // if queue is empty, queue stays empty;
          // otherwise, first read datum, then write datum
          if (~ac_token_queue_empty) begin
            // read => shift
            ac_tokens <= ac_tokens >> 1;
            ac_tokens_valid <= ac_tokens_valid >> 1;
            // write
            empty_bit_position = 0;
            while (ac_tokens_valid[empty_bit_position]) begin
              empty_bit_position = empty_bit_position + 1;
            end
            ac_tokens[empty_bit_position] <= ac_token_queue_in;
            ac_tokens_valid[empty_bit_position] <= 1'b1;
          end
        end
      end
    end
    endgenerate
`endif

endmodule
