// E.I.S. / ESA modlib for COMRADE and HardScale
  //
  // Description:
  // ----------------
  // Memory Write Modul with output queue and optional shift register mechanism.
  //
  // TODO: Parameter QDEPTH currently controls the size of the (undocumented) address/data input
  //       queue "queue_in" _and_ the size of the processed elements counter "pe_counter".
  //       To be able to use a processed elements counter _without_ an address/data
  //       input queue, there should be an additional parameter IN_QDEPTH which controls
  //       the address/data input queue size separately.
  //
  // Version History:
  // ----------------
  // 100826: BT: intial release


`timescale 1ns / 1ns

module memwrite

          #(parameter PORTNUM        = 0,
                      WIDTH          = 2'b00, // 00 = 32 bit, 10 = 16 bit, 01 = 8 bit
                      DEPTH          = 0,
                      QDEPTH         = 1,
                      TQDEPTH        = 1,
                      ADDR_WIDTH     = 24,
                      DATA_WIDTH     = 32,
                      RESULT_WIDTH   = 0,
                      STATIC_CT      = 0,
                      START_CTRL_IN    = 1,
                      DATA_SPECULATION = 0,
                      SPECULATION_BOUNDARY = 0,
                      PSEUDO_AT        = 0,  // WARNING: PSEUDO_AT functionality not yet implemented for memwrite! => always set PSEUDO_AT=0
                      NUM_CACHEPORTS = 1)


          (input  wire [ADDR_WIDTH-1:0]      A,
           input  wire [DATA_WIDTH-1:0]      B,
           input  wire                       RESET,
           input  wire                       CLK,
           input  wire                       CE,
           input  wire                       CONTROLLER_CANCEL,
           output wire                       TOKENS_EMPTY,// COCOMA needs to know if there are still tokens which have to be processed
           output wire                       CURRENT_TOKEN,
           output wire                       CURRENT_TOKEN_PSEUDO,
           input  wire                       START,
           output wire                       START_ACK,
           input  wire                       START_CTRL,
           output wire                       START_CTRL_ACK,
           input  wire                       START_CTRL_PSEUDO,
           input  wire                       CANCEL,
           output reg                        CANCEL_ACK,
           output wire                       CANCEL_STATE,
           input  wire                       CANCEL_STATE_ACK,
           output wire                       CANCEL_STATE_CTRL,
           input  wire                       CANCEL_STATE_CTRL_ACK,
           output wire                       RESULT_READY,            // COCOMA needs to know when store is finished
           input  wire                       RESULT_READY_ACK,

           output wire [(NUM_CACHEPORTS*ADDR_WIDTH)-1:0] CACHE_ADDR,
           output wire [(NUM_CACHEPORTS*2)-1:0]          CACHE_WIDTH,
           output wire [(NUM_CACHEPORTS*DATA_WIDTH)-1:0] CACHE_WRITE,
           output wire [NUM_CACHEPORTS-1:0]              CACHE_WE,
           input  wire [NUM_CACHEPORTS-1:0]              CACHE_STALL
           );


  // QDEPTH_INT has to be at least 16
  localparam QDEPTH_INT       = min_16(QDEPTH);
  // single CANCEL BIT since only valid transfers are allowed to enter
  // CANCEL BITS depending on LATENCY and shift register DEPTH
  localparam CANCEL_BITS      =  min_1(ceil_log2(QDEPTH_INT-1));
  // set QBITS accordingly
  localparam QBITS            = (QDEPTH == 0) ? 0 : ceil_log2(QDEPTH_INT-1);
  // TQDEPTH_INT has to be at least 16
  localparam TQDEPTH_INT      = min_16(TQDEPTH);
  // set TQBITS accordingly
  localparam TQBITS           = (TQDEPTH == 0) ? 0 : ceil_log2(TQDEPTH_INT-1);
  localparam QBITS_PE         = (QDEPTH == 0) ? 1 : QBITS;
  // COMPUTATION_BITS depending  on LATENCY shift register DEPTH and queue QDEPTH
  localparam COMPUTATION_BITS = (QDEPTH == 0) ? 1 : min_1(ceil_log2(QDEPTH_INT-1));


  wire ac_token_queue_in;
  wire pseudo_token_queue_in;
  wire ac_token_queue_out;
  wire ac_token_queue_out_pseudo;
  wire data_write;
  wire data_read;
  wire data_full;
  wire data_empty;
  wire access_completing;
  wire write_pe;
  wire read_pe;
  wire ac_token_queue_empty;
  wire pseudo_token_queue_empty;
  wire ac_token_queue_full;
  wire pseudo_token_queue_full;
  wire ac_token_queue_read;
  wire ac_token_queue_write;
  wire clear_access;
  wire OPERATION;

  wire [(ADDR_WIDTH + DATA_WIDTH - 1):0] DIN_Q;
  wire [(ADDR_WIDTH + DATA_WIDTH - 1):0] DOUT_Q;
  wire [(ADDR_WIDTH - 1)             :0] ADDR_Q;
  wire [(DATA_WIDTH - 1)             :0] DATA_Q;
        
  // a result is ready and consumed
  wire RESULT_CONSUMED = RESULT_READY & RESULT_READY_ACK;
  // a cancel request was accepted
  wire CANCEL_CONSUMED = CANCEL & CANCEL_ACK;
  
  // processed element counter
  reg[(QBITS_PE-1):0] pe_counter;
  reg  access_done;
  reg  full_pe;
  reg  empty_pe;
  reg  free_register;
  reg  cancel_register;

  reg  cancel_state_keep;
  reg  cancel_state_ctrl_keep;
  reg  cancel_state_incomplete;

`ifdef synthesis
`else
  reg  [COMPUTATION_BITS:0] computation_counter;
  reg  [CANCEL_BITS:0]      cancel_counter;
`endif

  // there is a valid result at the output, if there is a token in the processed elements queue
  // or an access is just finishing
  assign RESULT_READY = ~empty_pe | access_done;

  // reduce computation count by one
  assign clear_access = read_pe | (data_empty & START & START_ACK & (ac_token_queue_out & ~ac_token_queue_empty)) | (~data_empty & ((ac_token_queue_out & ~ac_token_queue_empty)));

  generate
  if (STATIC_CT == 1 & QDEPTH_INT == 0) begin
    // keep operation on stall
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        cancel_register <= 1'b0;
      end
      else begin
        if (CANCEL_CONSUMED) begin
	        cancel_register <= 1'b1;
	      end
	      else if (START & START_ACK) begin
	        cancel_register <= 1'b0;
	       end
	    end
	  end
	end

  if (STATIC_CT == 0 & QDEPTH_INT == 0) begin
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        cancel_register <= 1'b0;
      end
      else begin
         if (CANCEL_CONSUMED) begin
	         cancel_register <= 1'b1;
	       end
	       else if ((START & START_ACK) | (CANCEL_STATE & CANCEL_STATE_ACK)) begin
	         cancel_register <= 1'b0;
	       end
	    end
	  end
	end

  if (QDEPTH_INT == 0) begin
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        free_register   <= 1'b1;
      end
      else begin
        if (START & START_ACK & START_CTRL & START_CTRL_ACK) begin
          free_register <= 1'b0;
        end
        else if (RESULT_CONSUMED) begin
          free_register <= 1'b1;
        end
      end
    end
  end

  if (QDEPTH_INT == 0) begin
    assign START_CTRL_ACK = START & START_ACK & START_CTRL;
    assign START_ACK = START & ((START_CTRL & ~CACHE_STALL[PORTNUM] & free_register) | cancel_register);
    // when the operands and the AT are available, the operation can be started
    assign OPERATION = (START & START_ACK & START_CTRL & START_CTRL_ACK);
  end

  if (QDEPTH_INT > 0) begin
    // set when there is space left in the activate/cancel token queue
    assign START_CTRL_ACK = START_CTRL & (~ac_token_queue_full);
    // when the operands and the AT are available, the operation can be started
    assign OPERATION = (~data_empty | (START & START_ACK)) && ((START_CTRL & START_CTRL_ACK & ac_token_queue_empty) | (~ac_token_queue_out & ~ac_token_queue_empty));
  end
  endgenerate

  // ctrl_cmd == 1 means CANCEL, ctrl_cmd == 0 means ACTIVATE
  assign ac_token_queue_in = CANCEL_CONSUMED;

  // ac_token_queue_write validates ac_token_queue_in on token arrival
  assign ac_token_queue_write  = CANCEL_CONSUMED || (START_CTRL & START_CTRL_ACK);

  generate
  if (STATIC_CT == 1) begin
    // CANCEL_STATE_CTRL = 1, if there is a CT stored in this node
    //assign CANCEL_STATE_CTRL  = (((~ac_token_queue_empty & ac_token_queue_out)) & ~cancel_state_incomplete) | (cancel_state_ctrl_keep);
    assign CANCEL_STATE_CTRL  = 1'b0;
    // delete token when corresponding element is removed from queue
    // or whenever data is transferred or cancelled
    // ac_token_queue_read is executed whenever ac_token_queue_write is active and the data should not be written
    assign ac_token_queue_read = data_read | (data_empty & START & START_ACK & ~CACHE_STALL[PORTNUM] & ~ac_token_queue_empty & ~ac_token_queue_out)
                                | (data_empty & START & START_ACK & ~CACHE_STALL[PORTNUM] &  ac_token_queue_empty & START_CTRL & START_CTRL_ACK)
                                | (data_empty & START & START_ACK & ~ac_token_queue_empty & ac_token_queue_out);
  end

  // a read request is accepted, if the buffers are not full
  // there is no CANCEL by definition of the controller
  if (STATIC_CT == 1 & QDEPTH_INT > 0) begin
    assign START_ACK = START & ~data_full & ~full_pe;
    // CANCEL_STATE = 1, if there is a CT stored in this node
    assign CANCEL_STATE       = (((~ac_token_queue_empty & ac_token_queue_out)) & ~cancel_state_incomplete) | (cancel_state_keep);
  end
  if (STATIC_CT == 1 & QDEPTH_INT == 0) begin
    assign CANCEL_STATE       = (cancel_register & free_register);
  end

  if (STATIC_CT == 0) begin
    // CANCEL_STATE_CTRL is always zero
    assign CANCEL_STATE_CTRL = 1'b0;
    // delete token when corresponding element is removed from queue
    // or whenever data is transferred or cancelled
    // ac_token_queue_read is executed whenever ac_token_queue_write is active and the data should not be written
    assign ac_token_queue_read    = data_read | (data_empty & START & START_ACK & ~CACHE_STALL[PORTNUM] & ~ac_token_queue_empty & ~ac_token_queue_out)
                               | (data_empty & START & START_ACK & ~CACHE_STALL[PORTNUM] &  ac_token_queue_empty & START_CTRL & START_CTRL_ACK)
                               | (CANCEL_STATE & CANCEL_STATE_ACK);
  end

  // a read request is accepted, when the buffers are not full
  // do not accept data when CANCEL STATE is sent
  if (STATIC_CT == 0 & QDEPTH_INT > 0) begin
    assign START_ACK = START & ~data_full &  ~full_pe & ~CANCEL_STATE;
    // there is a cancel in the ac_queue but no corresponding element to cancel
    assign CANCEL_STATE = (ac_token_queue_out & ~ac_token_queue_empty) & data_empty;
  end

  if (STATIC_CT == 0 & QDEPTH_INT == 0) begin
    assign CANCEL_STATE       = (cancel_register & free_register);
  end

  endgenerate



  // assert CANCEL_STATE on CANCEL request
  // in case no computation is ongoin
  // deassert CANCEL_STATE on CANCEL_STATE_ACK input

  generate

  if (START_CTRL_IN == 0 & STATIC_CT == 0) begin
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
        CANCEL_ACK <= 1'b1;
    end
    else begin
      $display("ERROR: START_CTRL_IN has to be 1 for memread");
      $finish;
    end
  end
  end

  // no cancel and thus no cancel accept
  if (START_CTRL_IN == 0 & STATIC_CT == 1) begin
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
        CANCEL_ACK <= 1'b0;
    end
    else begin
      $display("ERROR: START_CTRL_IN has to be 1 for memread");
      $finish;
    end
  end
  end

  if (START_CTRL_IN == 1 & QDEPTH_INT > 0) begin
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
        CANCEL_ACK <= 1'b1;
    end
    else begin
        CANCEL_ACK <= ~ac_token_queue_full;
    end
  end
  end

  if (START_CTRL_IN == 1 & QDEPTH_INT == 0) begin
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
        CANCEL_ACK           <= 1'b1;
    end
    else begin
        CANCEL_ACK <= ~cancel_register;
    end
  end
  end


  if (START_CTRL_IN == 1 & STATIC_CT == 0) begin
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
        cancel_state_ctrl_keep  <= 1'b0;
        cancel_state_incomplete <= 1'b0;
    end
    else begin
      if (CANCEL_STATE_CTRL) begin
        // reset keep signal on handshake completion
        cancel_state_ctrl_keep <= ~CANCEL_STATE_CTRL_ACK;
        // initially keep signals are zero, thus a read determines whether there is something to process
        // as long as the keep signal is set an initial zero will remain zero, however an initial one can be set to zero
        // when keep goes low the condition is not fullfilled anymore
        cancel_state_incomplete <= (cancel_state_incomplete | ~(cancel_state_ctrl_keep)) & ~ac_token_queue_read;
        CANCEL_ACK           <= 1'b0;
      end
      else if  (cancel_state_incomplete & ac_token_queue_read) begin
          cancel_state_incomplete  <= 1'b0;
      end
    end
  end
  end


  if (START_CTRL_IN == 1 & STATIC_CT == 1) begin
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
        cancel_state_keep       <= 1'b0;
        cancel_state_ctrl_keep  <= 1'b0;
        cancel_state_incomplete <= 1'b0;
    end
    else begin
        if (CANCEL_STATE) begin
          // reset keep signal on handshake completion
          cancel_state_keep      <= ~CANCEL_STATE_ACK;
        end
        if (CANCEL_STATE_CTRL) begin
          // reset keep signal on handshake completion
          cancel_state_ctrl_keep <= ~CANCEL_STATE_CTRL_ACK;
        end

        if (CANCEL_STATE | CANCEL_STATE_CTRL) begin
          // initially keep signals are zero, thus a read determines whether there is something to process
          // as long as the keep signal is set an initial zero will remain zero, however an initial one can be set to zero
	        // when keep goes low the condition is not fullfilled anymore
          cancel_state_incomplete <= (cancel_state_incomplete | ~(cancel_state_keep | cancel_state_ctrl_keep)) & ~ac_token_queue_read;
        end // on ac_token_queue_read cancel has been processed
        else if (cancel_state_incomplete & ac_token_queue_read) begin
            cancel_state_incomplete  <= 1'b0;
        end
    end
  end
  end
  endgenerate

`ifdef synthesis
`else
  // visualization counters

  wire decrease_cancel_counter;
  wire increase_cancel_counter;
  wire decrease_computation_counter;
  wire increase_computation_counter;

  // count ongoing computations
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      computation_counter <=  {(COMPUTATION_BITS+1){1'b0}};
    end else begin
      case ({decrease_computation_counter, increase_computation_counter})
           2'b00: computation_counter <= computation_counter;
           2'b01: if (computation_counter != {1'b1, {(COMPUTATION_BITS){1'b0}}}) // max value for computation_counter
                 computation_counter <= computation_counter + 1'b1;
           2'b10: if (computation_counter != {(COMPUTATION_BITS+1){1'b0}})
                 computation_counter <= computation_counter - 1'b1;
           2'b11: computation_counter <= computation_counter;
      endcase
    end
  end

  generate
  if (STATIC_CT == 0) begin
    assign increase_computation_counter = START & START_ACK;
    assign increase_cancel_counter      = CANCEL_CONSUMED;
  end

  if (STATIC_CT == 1) begin
    assign increase_computation_counter = (START & START_ACK) && ~(~ac_token_queue_empty & data_empty & ac_token_queue_out);
    assign increase_cancel_counter      = CANCEL_CONSUMED;
  end
  endgenerate


  generate
  if (STATIC_CT == 0 & QDEPTH_INT > 0) begin
    assign decrease_computation_counter = clear_access;
    assign decrease_cancel_counter      = ac_token_queue_read & ac_token_queue_out & ~ac_token_queue_empty;
  end
  if (STATIC_CT == 0 & QDEPTH_INT == 0) begin
    assign decrease_computation_counter = clear_access | CANCEL_CONSUMED;
    assign decrease_cancel_counter      = CANCEL_CONSUMED;
  end
  if (STATIC_CT == 1 & QDEPTH_INT == 0) begin
    assign decrease_computation_counter = clear_access| CANCEL_CONSUMED;
    assign decrease_cancel_counter      = CANCEL_CONSUMED;
  end
  if (STATIC_CT == 1 & QDEPTH_INT > 0) begin
    assign decrease_computation_counter = clear_access;
    assign decrease_cancel_counter      = ac_token_queue_read & ((ac_token_queue_out & ~ac_token_queue_empty));
  end
  endgenerate


  // count non-extinguished CTs
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      cancel_counter <=  {(CANCEL_BITS+1){1'b0}};
    end else begin
        // decrease: delete address from address queue or from input
        // increase: CT comes in
        case ({decrease_cancel_counter, increase_cancel_counter})
             2'b00: cancel_counter <= cancel_counter;
             2'b01: if (cancel_counter != {1'b1, {(CANCEL_BITS){1'b0}}}) // max value for cancel_counter
                   cancel_counter <= cancel_counter + 1'b1;
             2'b10: if (cancel_counter != {(CANCEL_BITS+1){1'b0}})
                   cancel_counter <= cancel_counter - 1'b1;
             2'b11: cancel_counter <= cancel_counter;
        endcase
    end
  end
`endif

  // MARC completes a write access at the next posedge of CLK, if CACHE_WE is set now and there is no cache stall currently
  assign access_completing = CACHE_WE[PORTNUM] & ~CACHE_STALL[PORTNUM];

  // an access is finished one cycle after the access, such that successor accesses do not assign their
  // own request on the cache buses to early (this would interfere with the current access of this module)
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      access_done <= 1'b0;
    end else begin
      access_done <= access_completing;
    end
  end

  // write address/data to queue, if a new request is not directly consumed by MARC
  // and is not removed by a CT in the activate cancel token queue
  assign data_write = (START & START_ACK) && ~(~ac_token_queue_empty & data_empty & ac_token_queue_out)
                                     && ~(~ac_token_queue_empty & data_empty & ~ac_token_queue_out & ~CACHE_STALL[PORTNUM])
                                     && ~( ac_token_queue_empty & data_empty & START_CTRL & START_CTRL_ACK & ~CACHE_STALL[PORTNUM]);

  // read (= remove) address/data from address/data queue when either a request which layed on the queue has been passed to MARC
  // or an element is cancelled;
  assign data_read = ~data_empty & (access_completing  | (ac_token_queue_out & ~ac_token_queue_empty));

  // write a finished AT to the processed element queue, if an access is done, but there is no RESULT_READY_ACK yet
  // or there are already AT in the queue
  assign write_pe = access_completing & (~RESULT_CONSUMED | ~empty_pe);

  // remove an AT from the token queue as soon as there is a RESULT_READY_ACK
  assign read_pe = RESULT_CONSUMED;

  assign DIN_Q   = {A, B};
  assign ADDR_Q  = DOUT_Q[(ADDR_WIDTH + DATA_WIDTH - 1):(DATA_WIDTH)];
  assign DATA_Q  = DOUT_Q[(DATA_WIDTH - 1):0];

  // output assignment - set unused ports to 0 to prevent 'z' or 'x' values on the cache buses
  generate
  // NUM_CACHEPORTS-1 .. PORTNUM+1
  if (PORTNUM < NUM_CACHEPORTS - 1) begin
    assign CACHE_ADDR [NUM_CACHEPORTS*ADDR_WIDTH-1:(PORTNUM+1)*ADDR_WIDTH] = 0;
    assign CACHE_WRITE[NUM_CACHEPORTS*DATA_WIDTH-1:(PORTNUM+1)*DATA_WIDTH] = 0;
    assign CACHE_WIDTH[NUM_CACHEPORTS*2         -1:(PORTNUM+1)*2         ] = 0;
    assign CACHE_WE   [NUM_CACHEPORTS           -1:(PORTNUM+1)           ] = 0;
  end
  // PORTNUM (actual assignments)
  assign CACHE_ADDR [(PORTNUM+1)*ADDR_WIDTH-1:PORTNUM*ADDR_WIDTH] = OPERATION ? (data_empty ? A : ADDR_Q) : 0;
  assign CACHE_WRITE[(PORTNUM+1)*DATA_WIDTH-1:PORTNUM*DATA_WIDTH] = OPERATION ? (data_empty ? B : DATA_Q) : 0;
  assign CACHE_WIDTH[(PORTNUM+1)*2         -1:PORTNUM*2         ] = OPERATION ? WIDTH : 0;
  assign CACHE_WE   [ PORTNUM                                   ] = OPERATION;

  // PORTNUM-1 .. 0
  if (PORTNUM > 0) begin
    assign CACHE_ADDR [PORTNUM*ADDR_WIDTH-1:0] = 0;
    assign CACHE_WRITE[PORTNUM*DATA_WIDTH-1:0] = 0;
    assign CACHE_WIDTH[PORTNUM*2         -1:0] = 0;
    assign CACHE_WE   [PORTNUM           -1:0] = 0;
  end
  endgenerate


  // processed elements counter
  // stores activate tokens to signal completion
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      full_pe  <= 1'b0;
      empty_pe <= 1'b1;
      pe_counter <= {(QBITS_PE){1'b1}};
    end else begin

      if ((pe_counter[(QBITS_PE-1):0] == {{(QBITS_PE-1){1'b1}}, 1'b0}) & (write_pe & ~read_pe)) begin
        full_pe <= 1'b1;
      end
      else if ((pe_counter[(QBITS_PE-1):0] == {(QBITS_PE){1'b1}}) & (~read_pe | write_pe)) begin
        full_pe <= full_pe;
      end
      else begin
        full_pe <= 1'b0;
      end

      if ((pe_counter[(QBITS_PE-1):0] == {(QBITS_PE){1'b0}}) & (read_pe & ~write_pe)) begin
        empty_pe <= 1'b1;
      end
      else if ((pe_counter[(QBITS_PE-1):0] == {(QBITS_PE){1'b1}}) & (~write_pe | read_pe)) begin
        empty_pe <= empty_pe;
      end
      else begin
        empty_pe <= 1'b0;
      end

      if (write_pe == 1'b0 & read_pe == 1'b1 & empty_pe == 1'b0) begin
        pe_counter <= pe_counter - 1'b1;
      end
      // write_pe and no read_pe
      else if (write_pe == 1'b1 & read_pe == 1'b0 & full_pe == 1'b0) begin
        pe_counter <= pe_counter + 1'b1;
      end
      // simultaneous read / write_pe (or no read / write) into FIFO
      else begin
        pe_counter <= pe_counter;
      end

    end
  end

  // queue for addresses and data
  disram_queue_non_spec #(.WR(ADDR_WIDTH + DATA_WIDTH), .QDEPTH(QDEPTH_INT), .QBITS(QBITS)) queue_in (
    .CLK(CLK),
    .RESET(RESET),
    .DIN(DIN_Q),
    .READ(data_read),
    .WRITE(data_write),
    .DOUT(DOUT_Q),
    .EMPTY(data_empty),
    .FULL(data_full));

  // queue to keep activate and cancel tokens
  sr_queue #(.WR(1), .QDEPTH(QDEPTH_INT), .QBITS(QBITS)) ac_token_queue (
    .CLK(CLK),
    .RESET(RESET),
    .DIN(ac_token_queue_in),
    .READ(ac_token_queue_read),
    .WRITE(ac_token_queue_write),
    .DOUT(ac_token_queue_out),
    .EMPTY(ac_token_queue_empty),
    .FULL(ac_token_queue_full));

  generate
  if (PSEUDO_AT == 1) begin
  // pseudoAT queue stores a "1" per token of ac_token_queue if that token is a pseudo AT, a "0" otherwise
    sr_queue #(.WR(1), .QDEPTH(QDEPTH_INT), .QBITS(QBITS)) ac_token_queue_pseudo (
      .CLK(CLK),
      .RESET(RESET),
      .DIN(pseudo_token_queue_in),
      .READ(ac_token_queue_read),
      .WRITE(ac_token_queue_write),
      .DOUT(ac_token_queue_out_pseudo),
      .EMPTY(pseudo_token_queue_empty),
      .FULL(pseudo_token_queue_full));
  end
  endgenerate

  // ----- TOKENS_EMPTY -----
  generate
  if (START_CTRL_IN == 1) begin
    assign TOKENS_EMPTY = ac_token_queue_empty;
  end
  if (START_CTRL_IN == 0) begin
    assign TOKENS_EMPTY = 1'b1;
  end
  endgenerate

  // ----- CURRENT_TOKEN -----
  generate
  if (START_CTRL_IN == 1) begin
    assign CURRENT_TOKEN = ~ac_token_queue_empty & ac_token_queue_out; // ac_token_queue_out means CT, ~ac_token_queue_out means AT => CURRENT_TOKEN means CT, ~CURRENT_TOKEN means AT
  end
  if (START_CTRL_IN == 0) begin
    assign CURRENT_TOKEN = 1'b0; // no token queue => no information about stored tokens
  end
  endgenerate

  // ----- CURRENT_TOKEN_PSEUDO -----
  generate
  if (START_CTRL_IN == 1) begin
    assign CURRENT_TOKEN_PSEUDO = ~ac_token_queue_empty & ac_token_queue_out_pseudo;
  end
  if (START_CTRL_IN == 0) begin
    assign CURRENT_TOKEN_PSEUDO = 1'b0; // no token queue => no information about stored tokens
  end
  endgenerate

  `include "functions.v"


  // make state of ac_token_queue visible during simulation
`ifdef synthesis
`else
reg [15:0] ac_tokens_valid;
reg [15:0] ac_tokens;
integer empty_bit_position;
generate
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        ac_tokens_valid <= 0;
        ac_tokens <= 0;
        empty_bit_position = 0;
      end else begin
        if (ac_token_queue_write & ~ac_token_queue_read) begin
          // write to queue, insert data at empty bit position, start to search at LSB
          empty_bit_position = 0;
          while (ac_tokens_valid[empty_bit_position]) begin
            empty_bit_position = empty_bit_position + 1;
          end
          ac_tokens[empty_bit_position] <= ac_token_queue_in;
          ac_tokens_valid[empty_bit_position] <= 1'b1;
        end else if (~ac_token_queue_write & ac_token_queue_read) begin
          // read from queue, delete data at bit position 0 by right-shifting
          ac_tokens <= ac_tokens >> 1;
          ac_tokens_valid <= ac_tokens_valid >> 1;
        end else if (~ac_token_queue_write & ~ac_token_queue_read) begin
          // do nothing
        end else if (ac_token_queue_write & ac_token_queue_read) begin
          // if queue is empty, queue stays empty;
          // otherwise, first read datum, then write datum
          if (~ac_token_queue_empty) begin
            // read => shift
            ac_tokens <= ac_tokens >> 1;
            ac_tokens_valid <= ac_tokens_valid >> 1;
            // write
            empty_bit_position = 0;
            while (ac_tokens_valid[empty_bit_position]) begin
              empty_bit_position = empty_bit_position + 1;
            end
            ac_tokens[empty_bit_position] <= ac_token_queue_in;
            ac_tokens_valid[empty_bit_position] <= 1'b1;
          end
        end
      end
    end
endgenerate
`endif

endmodule
