// E.I.S. / ESA modlib for COMRADE and HardScale
  //
  //// Description:
  // ----------------
  // Universal floating point multiplication
  // with output queue and shift register mechanism
  //
  // Version History:
  // ----------------
  // 100826: BT: intial release

`timescale 1ns / 1ns

module mulfloat
         #(parameter
                      WA                   = 32,
                      WB                   = 32,
                      WR                   = 32,
                      DEPTH                = 0,
                      QDEPTH               = 16,
                      TQDEPTH              = 16,
                      STATIC_CT            = 0,
                      START_CTRL_IN        = 0,
                      DATA_SPECULATION     = 0,
                      SPECULATION_BOUNDARY = 0,
                      PSEUDO_AT            = 0,
                      SIGN                 = 0
          )(
           input  wire    [(WA-1):0] A,
           input  wire    [(WB-1):0] B,
           output wire    [(WR-1):0] R,
           input  wire               RESET,
           input  wire               CLK,
           input  wire               CE,
           input  wire               CONTROLLER_CANCEL,
           output wire               TOKENS_EMPTY,
           output wire               CURRENT_TOKEN,
           output wire               CURRENT_TOKEN_PSEUDO,
           input  wire               START,
           output wire               START_ACK,
           input  wire               START_CTRL,
           output wire               START_CTRL_ACK,
           input  wire               START_CTRL_PSEUDO,
           input  wire               CANCEL,
           output reg                CANCEL_ACK,
           output wire               CANCEL_STATE,
           input  wire               CANCEL_STATE_ACK,
           output wire               CANCEL_STATE_CTRL,
           input  wire               CANCEL_STATE_CTRL_ACK,
           output wire               RESULT_READY,
           input  wire               RESULT_READY_ACK);

  // define LATENCY of this module - high speed implementation using 3 / 10 DSP48E Blocks
  localparam LATENCY          = (WA == 32) ? 8 : 10;
  // CANCEL BITS depending on LATENCY and shift register DEPTH
  localparam CANCEL_BITS      = min_1(ceil_log2(LATENCY+DEPTH-1));
  // QDEPTH_INT has to be at least 16
  localparam QDEPTH_INT       = min_16(QDEPTH);
  // set QBITS accordingly
  localparam QBITS            = (QDEPTH == 0) ? 0 : ceil_log2(QDEPTH_INT-1);
  // TQDEPTH_INT has to be at least 16
  localparam TQDEPTH_INT      = min_16(TQDEPTH);
  // set TQBITS accordingly
  localparam TQBITS           = (TQDEPTH == 0) ? 0 : ceil_log2(TQDEPTH_INT-1);
  // COMPUTATION_BITS depending  on LATENCY shift register DEPTH and queue QDEPTH
  localparam COMPUTATION_BITS = min_1(ceil_log2(LATENCY+DEPTH+QDEPTH_INT-1));

  wire [(WR-1):0]       prod;


  `include "handshaking_assigns_blackbox_include.v"

  initial begin
    if (((WA != 32) && (WA != 64)) || (WB != WA) || (WR != WA)) begin
      $write("Module MULFLOAT: Currently supports only a width of 64b/32b for A, B, R");
      $finish;
    end
  end

  generate
  if (WA == 32) begin
    mul_f_32_32 mulf32 (
      .a(A),           // Bus [31 : 0]
      .b(B),           // Bus [31 : 0]
      .result(prod),   // Bus [31 : 0]
      .operation_rfd(rfd),
      .operation_nd(START),
      .rdy(rdy),
      .clk(CLK),
      .sclr(RESET),
      .ce(CE_IP_BLOCK));
  end
  else if (WA == 64) begin
    mul_f_64_64 divf64 (
      .a(A),           // Bus [63 : 0]
      .b(B),           // Bus [63 : 0]
      .result(prod),   // Bus [63 : 0]
      .operation_rfd(rfd),
      .operation_nd(START),
      .rdy(rdy),
      .clk(CLK),
      .sclr(RESET),
      .ce(CE_IP_BLOCK));
  end
  endgenerate

  // shift register instantiation
  shiftreg #(.WA(WR), .WR(WR), .DEPTH(DEPTH)) sr (
    .A(prod),
    .R(DATA_SHIFTREG_TO_QUEUE),
    .RESET(RESET),
    .CLK(CLK),
    .CE(CE_SHIFTREG),
    .START(rdy),
    .RESULT_READY(RESULT_READY_SHIFTREG));



  `include "functions.v"

endmodule
