// E.I.S. / ESA modlib for COMRADE and HardScale
  //
  // Version History:
  // ----------------
  // 100826: BT: intial release

`timescale 1ns / 1ns

module outreg
         #(parameter
                      WA                    = 32,
                      WR                    = WA,
                      SIGN                  = 0, // dummy
                      IOADDR                = 2,
                      IOAWIDTH              = 4,
                      STATIC_CT             = 0,
                      START_CTRL_IN         = 0,
                      DATA_SPECULATION      = 0,
                      SPECULATION_BOUNDARY  = 0,
                      DEPTH                 = 0, // dummy
                      QDEPTH                = 0,
                      TQDEPTH               = 0,
                      PSEUDO_AT             = 0
          )(
           input  wire [(WA-1):0]       A,
           output reg  [(WR-1):0]       R,
           input  wire                  RESET,
           input  wire                  CLK,
           input  wire                  CE,
           input  wire                  CONTROLLER_CANCEL,
           output wire                  TOKENS_EMPTY,
           output wire                  CURRENT_TOKEN,
           output wire                  CURRENT_TOKEN_PSEUDO,
           input  wire [(IOAWIDTH-1):0] IOA,
           input  wire                  IOSEL,
           output wire [(WR-1):0]       IODOUT,
           input  wire                  START,
           output wire                  START_ACK,
           input  wire                  START_CTRL,
           output wire                  START_CTRL_ACK,
           input  wire                  START_CTRL_PSEUDO,
           input  wire                  CANCEL,
           output wire                  CANCEL_ACK,
           output reg                   CANCEL_STATE,
           input  wire                  CANCEL_STATE_ACK,
           output reg                   CANCEL_STATE_CTRL,
           input  wire                  CANCEL_STATE_CTRL_ACK,
           output reg                   RESULT_READY,
           input  wire                  RESULT_READY_ACK);


  wire  computation_counter = 1'b0; // for PDF counter obersvation only
  wire  cancel_counter = 1'b0;      // for PDF counter obersvation only
  
  wire  ac_token_queue_in;
  wire  ac_token_queue_out;
  wire  ac_token_queue_empty;
  wire  ac_token_queue_full;
  wire  ac_token_queue_read;
  wire  ac_token_queue_write;
  wire  remove_ac_at;
  wire  remove_ac_ct;


  // this must be combined with all other OUTREGS using a wide-OR.
  assign IODOUT = (IOSEL && (IOA == IOADDR)) ? R : 0;


  // R
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      R <= 0;
    end
    else begin
      // accept data from HW Ops if enabled, newdata, and AT (if control edge exists)
      if (CE) begin
        if (START & START_ACK & ((START_CTRL_IN == 0) || (~ac_token_queue_empty & ~ac_token_queue_out))) begin
          R <= A;
        end
      end
    end
  end

  // RESULT_READY
  generate
  if (START_CTRL_IN == 1) begin
    // control edge
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        RESULT_READY <= 0;
      end else begin
        if (CE) begin
          // accept data only if AT is in token queue, and incoming data is not deleted by pending CT
          if (((START & START_ACK)) & (~ac_token_queue_empty & ~ac_token_queue_out & ~CANCEL_STATE)) begin
            RESULT_READY <= 1;
          end else if (RESULT_READY & RESULT_READY_ACK) begin
            RESULT_READY <= 0;
          end
        end
      end
    end
  end
  if (START_CTRL_IN == 0) begin
    // no control edge => no token relevance
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        RESULT_READY <= 0;
      end else begin
        if (CE) begin
          if (START & START_ACK) begin
            RESULT_READY <= 1;
          end else if (RESULT_READY & RESULT_READY_ACK) begin
            RESULT_READY <= 0;
          end
        end
      end
    end
  end
  endgenerate

  // START_CTRL_ACK
  generate
  if (START_CTRL_IN == 1) begin
    // activate/cancel queue required, set when there is space left in the token queue
    assign START_CTRL_ACK = CE & START_CTRL & ~ac_token_queue_full;
  end
  if (START_CTRL_IN == 0) begin
    assign START_CTRL_ACK = 1'b0;
  end
  endgenerate

  // START_ACK
  generate
  if (START_CTRL_IN == 1 & STATIC_CT == 1) begin
      // static CTs:
      // accept new data, if there is an AT in the queue and there is currently no output data, or
      // there is a CT pending
      assign START_ACK = CE & START & ((~ac_token_queue_empty & ~ac_token_queue_out & ~RESULT_READY) |
                                       CANCEL_STATE);
  end

  if (START_CTRL_IN == 1 & STATIC_CT == 0) begin
      // dynamic CTs:
      // accept new data, if there is an AT in the queue, and there is currently no output data
      assign START_ACK = CE & START & ~ac_token_queue_empty & ~ac_token_queue_out & ~RESULT_READY;
  end
  if (START_CTRL_IN == 0) begin
    assign START_ACK = CE & ~RESULT_READY;
  end
  endgenerate

  // CANCEL_STATE
  generate
  if (START_CTRL_IN == 1 & STATIC_CT == 1) begin
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        CANCEL_STATE <= 1'b0;
      end
      else begin
          CANCEL_STATE <= (CANCEL_STATE & START & START_ACK) ? 0 // reset CANCEL_STATE on START & START_ACK
                        : (~CANCEL_STATE & remove_ac_ct)     ? 1 // set CANCEL_STATE on remove_ac_ct
                        : CANCEL_STATE;
      end
    end
  end
  if (START_CTRL_IN == 1 & STATIC_CT == 0) begin
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        CANCEL_STATE <= 1'b0;
      end
      else begin
          CANCEL_STATE <= (CANCEL_STATE & CANCEL_STATE_ACK) ? 0 // reset CANCEL_STATE on CANCEL_STATE_ACK
                        : (~CANCEL_STATE & remove_ac_ct)      ? 1 // set CANCEL_STATE on remove_ac_ct
                        : CANCEL_STATE;
      end
    end
  end

  if (START_CTRL_IN == 0) begin
    always @(posedge CLK, posedge RESET) begin
      CANCEL_STATE <= 1'b0;
    end
  end

  endgenerate

  // CANCEL_STATE_CTRL
  generate
  if (START_CTRL_IN == 1) begin
    always @(posedge CLK, posedge RESET) begin
      if (RESET) begin
        CANCEL_STATE_CTRL <= 1'b0;
      end else begin
        CANCEL_STATE_CTRL <= (CANCEL_STATE_CTRL & CANCEL_STATE_CTRL_ACK) ? 0 // reset CANCEL_STATE_CTRL on CANCEL_STATE_CTRL_ACK
                           : (~CANCEL_STATE_CTRL & remove_ac_ct)           ? 1 // set CANCEL_STATE_CTRL on remove_ac_ct
                           : CANCEL_STATE_CTRL;
      end
    end
  end
  if (START_CTRL_IN == 0) begin
    always @(posedge CLK, posedge RESET) begin
      CANCEL_STATE_CTRL <= 1'b0;
    end
  end
  endgenerate

  // CANCEL_ACK
  generate
  if (START_CTRL_IN == 1) begin
    assign CANCEL_ACK = ~ac_token_queue_full;
  end
  if (START_CTRL_IN == 0) begin
    assign CANCEL_ACK = 1'b0;
  end
  endgenerate

  // ac_token_queue and associated signals
  generate
  if (START_CTRL_IN == 1) begin
    // ac_token_queue_in == 1 means CANCEL, ac_token_queue_in == 0 means ACTIVATE
    assign ac_token_queue_in = CE & CANCEL & CANCEL_ACK;

    // ac_token_queue_write validates ac_token_queue_in on token arrival
    assign ac_token_queue_write = CE & (CANCEL & CANCEL_ACK) | (START_CTRL & START_CTRL_ACK);

    // delete AT, when the queue is not empty, there is an AT in the queue, and there is data coming in
    // (i.e. the AT validates this data), and there is no pending CT
    assign remove_ac_at = CE & ~ac_token_queue_empty & ~ac_token_queue_out & START & START_ACK & ~CANCEL_STATE;

    // delete CT, when the queue is not empty, there is a CT in the queue,
    // and the CT is moved to CANCEL_STATE / CANCEL_STATE_CTRL
    assign remove_ac_ct = CE & ~ac_token_queue_empty & ac_token_queue_out & ~CANCEL_STATE & ~CANCEL_STATE_CTRL;// delete token when either at or ct is removed
    assign ac_token_queue_read = remove_ac_at | remove_ac_ct;
    sr_queue #(.WR(1), .QDEPTH(16), .QBITS(4)) ac_token_queue (
      .CLK(CLK),
      .RESET(RESET),
      .DIN(ac_token_queue_in),
      .READ(ac_token_queue_read),
      .WRITE(ac_token_queue_write),
      .DOUT(ac_token_queue_out),
      .EMPTY(ac_token_queue_empty),
      .FULL(ac_token_queue_full));
  end
  if (START_CTRL_IN == 0) begin
    assign ac_token_queue_in    = 1'b0;
    assign ac_token_queue_write = 1'b0;
    assign remove_ac_at         = 1'b0;
    assign remove_ac_ct         = 1'b0;
    assign ac_token_queue_read  = 1'b0;
  end
  endgenerate

  // TOKENS_EMPTY, CURRENT_TOKEN, CURRENT_TOKEN_PSEUDO
  generate
  if (START_CTRL_IN == 1) begin
    assign TOKENS_EMPTY = ~RESULT_READY & ac_token_queue_empty;
    assign CURRENT_TOKEN = ~ac_token_queue_empty & ac_token_queue_out;
  end else begin
    assign TOKENS_EMPTY = ~RESULT_READY;
    assign CURRENT_TOKEN = 1'b0;
  end
  assign CURRENT_TOKEN_PSEUDO = 1'b0;
  endgenerate

endmodule

