// E.I.S. / ESA modlib for COMRADE and HardScale
  //
  // Description:
  // ----------------
  // Output queue
  // based on shift registers
  //
  // Version History:
  // ----------------
  // 100826: BT: intial release


module sr_queue
         #(parameter
                      WR     = 32,
                      QDEPTH = 16,
                      QBITS  =  4  // 4 bit -> 16 words in the QUEUE.
          )(
           input                     CLK,
           input                     RESET,
           input  [(WR-1):0]         DIN,
           input                     READ,
           input                     WRITE,
           output wire [(WR-1):0]    DOUT,
           output reg                EMPTY,
           output reg                FULL);


 // QBITS==0 can only result from QDEPTH=0; in this case, ADDRESS_SRL
 // is not generated anyway, but XST does not recognize this, so
 // we have to guarantee that in [x:y] for ADDRESS_SRL, x is
 localparam QBITS_NZ = QBITS + (QBITS==0);

 reg  [(QBITS_NZ-1):0] ADDRESS_SRL;
 wire [(WR-1)   :0]            DOUT_SRLC;
 wire                          CE;

 assign DOUT = (QDEPTH == 0)              ? DIN  : DOUT_SRLC;
 assign CE   = (FULL & (~(READ & WRITE))) ? 1'b0 : WRITE; // generate the required FIFO
  genvar i;

  generate
  if (QDEPTH != 0) begin

   for (i=0; i < WR; i=i+1) begin: i_SRLCXXE

  // shift register instantiation
     srlcxxe #(.DEPTH(QDEPTH), .ADR_BITS(QBITS_NZ)) SRLCXXE_i (
      .D(DIN[i]),
      .Q(DOUT_SRLC[i]),
      .A(ADDRESS_SRL[(QBITS_NZ-1):0]),
      .CLK(CLK),
      .CE(CE));
   end
  end
  endgenerate


generate
if (QDEPTH != 0) begin// address pointer logic
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      ADDRESS_SRL <= {(QBITS_NZ){1'b1}};
    end else begin
      if (WRITE == 1'b0 & READ == 1'b1 & EMPTY == 1'b0) begin
              ADDRESS_SRL <= ADDRESS_SRL - 1'b1;
      end
      // write into FIFO
      else if (WRITE == 1'b1 & READ == 1'b0 & FULL == 1'b0) begin
            // increase pointer
             ADDRESS_SRL <= ADDRESS_SRL + 1'b1;
      end
      // simultaneous read / write (or no read / write) into FIFO
      else begin
          ADDRESS_SRL <= ADDRESS_SRL;
      end
    end
  end


  // Comparator Block
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      FULL <= 1'b0;
    end else begin
      if ((ADDRESS_SRL[(QBITS_NZ-1):0] == {{(QBITS_NZ-1){1'b1}}, 1'b0}) & (WRITE & ~READ)) begin
        FULL <= 1'b1;
      end
      else if ((ADDRESS_SRL[(QBITS_NZ-1):0] == {(QBITS_NZ){1'b1}}) & (~READ | WRITE)) begin
        FULL <= FULL;
      end
      else begin
        FULL <= 1'b0;
      end
    end
  end


  // Empty flag is generated when reading from the last location
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      EMPTY <= 1'b1;
    end else begin
      if ((ADDRESS_SRL[(QBITS_NZ-1):0] == {(QBITS_NZ){1'b0}}) & (READ & ~WRITE)) begin
        EMPTY <= 1'b1;
      end
      else if ((ADDRESS_SRL[(QBITS_NZ-1):0] == {(QBITS_NZ){1'b1}}) & (~WRITE | READ)) begin
        EMPTY <= EMPTY;
      end
      else begin
        EMPTY <= 1'b0;
      end
    end
  end

end


else begin
  // Comparator Block
  always @(posedge CLK) begin
    FULL  <= 1'b1;
    EMPTY <= 1'b1;
  end
end
endgenerate

endmodule
