// E.I.S. / ESA modlib for COMRADE and HardScale
  //
  //// Description:
  // ----------------
  // Module used for AT/CT token buffering (no data flow). Combinatoric behavior
  // possible (in case of empty token counter) and wanted.
  //
  // Version History:
  // ----------------
  // 100826: BT: intial release

`timescale 1ns / 1ns

module token_queue
         #(parameter
                      COUNTER_WIDTH = 3
          )(
           input  wire                   RESET,
           input  wire                   CLK,
           input  wire                   CE,
           input  wire                   START,
           output wire                   START_ACK,
           input  wire                   CANCEL,
           output wire                   CANCEL_ACK,
           output wire                   CANCEL_STATE,
           input  wire                   CANCEL_STATE_ACK,
           output wire                   CANCEL_STATE_CTRL,
           input  wire                   CANCEL_STATE_CTRL_ACK,
           output wire                   RESULT_READY,
           input  wire                   RESULT_READY_ACK);

  `include "functions.v"
  
  localparam COUNTER_WIDTH_INT = min_1(COUNTER_WIDTH);

  // count tokens
  reg [COUNTER_WIDTH_INT-1:0] TOKEN_COUNTER;

  // type of token (0 = CT, 1 = AT)
  reg ATnCT;

  wire  INCREASE_token_queue_COUNTER_AT;
  wire  DECREASE_token_queue_COUNTER_AT;
  wire  INCREASE_token_queue_COUNTER_CT;
  wire  DECREASE_token_queue_COUNTER_CT;
  wire  INCREASE_token_queue_COUNTER;
  wire  DECREASE_token_queue_COUNTER;

  wire  EMPTY;
  wire  FULL;

  wire  [(COUNTER_WIDTH_INT-1):0] cancel_counter = TOKEN_COUNTER & {(COUNTER_WIDTH_INT){~ATnCT}};
  wire  [(COUNTER_WIDTH_INT-1):0] computation_counter = TOKEN_COUNTER & {(COUNTER_WIDTH_INT){ATnCT}};


  assign EMPTY = TOKEN_COUNTER == {(COUNTER_WIDTH_INT){1'b0}};
  assign FULL = TOKEN_COUNTER == {(COUNTER_WIDTH_INT){1'b1}};

  // increase the counter (AT), if ATs are currently stored (or the counter is 0), and
  // an AT comes in via START and is acknowledged (START_ACK)
  assign INCREASE_token_queue_COUNTER_AT = (ATnCT | EMPTY) ? (START & START_ACK) : 0;

  // decrease the counter (AT), if an AT is consumed by the successors (RESULT_READY & RESULT_READY_ACK)
  assign DECREASE_token_queue_COUNTER_AT = RESULT_READY & RESULT_READY_ACK;

  // increase the counter (CT), if CTs are currently stored (or the counter is 0), and
  // a CT comes in and is acknowledged (CANCEL & CANCEL_ACK)
  assign INCREASE_token_queue_COUNTER_CT = (~ATnCT | EMPTY) ? (CANCEL & CANCEL_ACK) : 0;

  // decrease the counter (CT), if a CT is consumed by the successors (CANCEL_STATE & CANCEL_STATE_ACK)
  assign DECREASE_token_queue_COUNTER_CT = CANCEL_STATE & CANCEL_STATE_ACK;

  assign INCREASE_token_queue_COUNTER = INCREASE_token_queue_COUNTER_AT | INCREASE_token_queue_COUNTER_CT;
  assign DECREASE_token_queue_COUNTER = DECREASE_token_queue_COUNTER_AT | DECREASE_token_queue_COUNTER_CT;

  // accept an AT, if ATs are currently stored and the counter is not full
  assign START_ACK = START & (ATnCT | EMPTY) & ~FULL;

  // accept a CT, if CTs are currently stored and the counter is not full
  assign CANCEL_ACK = (~ATnCT | EMPTY) & ~FULL;

  // forward a CT if the counter contains CTs or the counter is empty and a CT comes in
  assign CANCEL_STATE = 1'b0;
  assign CANCEL_STATE_CTRL = (~ATnCT & ~EMPTY) | (EMPTY & CANCEL & CANCEL_ACK);

  // forward an AT if the counter contains ATs or the counter is empty and an AT comes in
  assign RESULT_READY = (ATnCT & ~EMPTY) | (EMPTY & START);


  // ATnCT
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      ATnCT <= 0;
    end else begin
      if (CE) begin
        if (INCREASE_token_queue_COUNTER_AT) begin
          ATnCT <= 1; // store ATs
        end else if (INCREASE_token_queue_COUNTER_CT) begin
          ATnCT <= 0; // store CTs
        end
      end
    end
  end

  // TOKEN_COUNTER
  always @(posedge CLK, posedge RESET) begin
    if (RESET) begin
      TOKEN_COUNTER <= 0;
    end else begin
      if (CE) begin
        case ({INCREASE_token_queue_COUNTER, DECREASE_token_queue_COUNTER})
          2'b00: TOKEN_COUNTER <= TOKEN_COUNTER;
          2'b01: if (~EMPTY) begin
                   TOKEN_COUNTER <= TOKEN_COUNTER - 1;
                 end
          2'b10: if (~FULL) begin
                   TOKEN_COUNTER <= TOKEN_COUNTER + 1;
                 end
          2'b11: TOKEN_COUNTER <= TOKEN_COUNTER;
        endcase
      end
    end
  end

endmodule
